"""
    Homonim: Correction of aerial and satellite imagery to surface reflectance
    Copyright (C) 2021 Dugal Harris
    Email: dugalh@gmail.com

    This program is free software: you can redistribute it and/or modify
    it under the terms of the GNU Affero General Public License as
    published by the Free Software Foundation, either version 3 of the
    License, or any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU Affero General Public License for more details.

    You should have received a copy of the GNU Affero General Public License
    along with this program.  If not, see <https://www.gnu.org/licenses/>.
"""
import os
from pathlib import Path
from typing import Tuple, Dict

import pytest
import rasterio as rio
import yaml
from pytest import FixtureRequest
from rasterio.features import shapes
from rasterio.vrt import WarpedVRT

from homonim import utils
from homonim.enums import ProcCrs, Model
from homonim.errors import IoError
from homonim.fuse import RasterFuse


@pytest.mark.parametrize(
    'src_file, ref_file', [
        ('src_file_50cm_float', 'ref_file_100cm_float'),
        ('src_file_100cm_float', 'ref_file_50cm_float'),
    ]
)  # yapf: disable
def test_creation(src_file: str, ref_file: str, tmp_path: Path, request: FixtureRequest):
    """ Test creation and configuration of RasterFuse. """
    src_file: Path = request.getfixturevalue(src_file)
    ref_file: Path = request.getfixturevalue(ref_file)
    raster_fuse = RasterFuse(src_file, ref_file)
    with raster_fuse:
        assert (raster_fuse.proc_crs != ProcCrs.auto)
        assert (not raster_fuse.closed)
    assert raster_fuse.closed


@pytest.mark.parametrize('overwrite', [False, True])
def test_overwrite(tmp_path: Path, src_file_50cm_float, ref_file_100cm_float, overwrite: bool):
    """ Test overwrite behaviour. """
    corr_filename = tmp_path.joinpath('corrected.tif')
    param_filename = utils.create_param_filename(corr_filename)
    params = dict(
        corr_filename=corr_filename, param_filename=param_filename, model=Model.gain_blk_offset, kernel_shape=(5, 5),
        overwrite=overwrite,
    )

    raster_fuse = RasterFuse(src_filename=src_file_50cm_float, ref_filename=ref_file_100cm_float)
    corr_filename.touch()
    with raster_fuse:
        if not overwrite:
            with pytest.raises(FileExistsError):
                raster_fuse.process(**params)
        else:
            raster_fuse.process(**params)

    os.remove(corr_filename)
    param_filename.touch()
    with raster_fuse:
        if not overwrite:
            with pytest.raises(FileExistsError):
                raster_fuse.process(**params)
        else:
            raster_fuse.process(**params)


@pytest.mark.parametrize(
    'src_file, ref_file, model, kernel_shape, max_block_mem', [
        ('src_file_45cm_float', 'ref_file_100cm_float', Model.gain, (1, 1), 2.e-4),
        ('src_file_45cm_float', 'ref_file_100cm_float', Model.gain_blk_offset, (1, 1), 1.e-3),
        ('src_file_45cm_float', 'ref_file_100cm_float', Model.gain_offset, (5, 5), 1.e-3),
        ('src_file_100cm_float', 'ref_file_45cm_float', Model.gain, (1, 1), 2.e-4),
        ('src_file_100cm_float', 'ref_file_45cm_float', Model.gain_blk_offset, (1, 1), 1.e-3),
        ('src_file_100cm_float', 'ref_file_45cm_float', Model.gain_offset, (5, 5), 1.e-3),
        ('src_file_45cm_float', 'ref_file_wgs84_sup_float', Model.gain_blk_offset, (1, 1), 1.e-3),
        ('src_file_wgs84_sup_100cm_float', 'ref_file_45cm_float', Model.gain_blk_offset, (1, 1), 1.e-3),
    ]
)  # yapf: disable
def test_basic_fusion(
    src_file: str, ref_file: str, model: Model, kernel_shape: Tuple[int, int], max_block_mem: float, tmp_path: Path,
    request: FixtureRequest,
):
    """ Test fusion output with different src/ref images, and model etc combinations. """
    src_file = request.getfixturevalue(src_file)
    ref_file = request.getfixturevalue(ref_file)
    block_config = RasterFuse.create_block_config(max_block_mem=max_block_mem)
    corr_filename = tmp_path.joinpath('corrected.tif')
    raster_fuse = RasterFuse(src_file, ref_file)
    with raster_fuse:
        raster_fuse.process(corr_filename, model, kernel_shape, block_config=block_config)
    assert (corr_filename.exists())
    # open src_file in a WarpedVRT to reproject it North-up (if necessary)
    with WarpedVRT(rio.open(src_file, 'r')) as src_ds, rio.open(corr_filename, 'r') as out_ds:
        src_array = src_ds.read(indexes=1)
        src_mask = src_ds.dataset_mask().astype('bool', copy=False)
        out_array = out_ds.read(indexes=1)
        out_mask = out_ds.dataset_mask().astype('bool', copy=False)
        assert (out_mask == src_mask).all()
        assert (out_array[out_mask] == pytest.approx(src_array[src_mask], abs=2))


@pytest.mark.parametrize(
    'out_profile', [
        dict(
            driver='GTiff', dtype='float32', nodata=float('nan'),
            creation_options=dict(
                tiled=True, blockxsize=512, blockysize=512, compress='deflate', interleave='band',
                #photometric='minisblack'
            )
        ),
        dict(
            driver='GTiff', dtype='uint8', nodata=0,
            creation_options=dict(
                tiled=True, blockxsize=64, blockysize=64, compress='jpeg', interleave='pixel', photometric='ycbcr'
            )
        ),
        dict(driver='PNG', dtype='uint16', nodata=0, creation_options=dict()),
    ]
)  # yapf: disable
def test_out_profile(file_rgb_100cm_float, tmp_path: Path, out_profile: Dict):
    """ Test fusion output image format (profile) with different out_profile configurations. """
    raster_fuse = RasterFuse(file_rgb_100cm_float, file_rgb_100cm_float)
    corr_filename = tmp_path.joinpath('corrected.tif')
    with raster_fuse:
        raster_fuse.process(corr_filename, Model.gain_blk_offset, (3, 3), out_profile=out_profile)
    assert (corr_filename.exists())
    out_profile.update(**out_profile['creation_options'])
    out_profile.pop('creation_options')
    with rio.open(file_rgb_100cm_float, 'r') as src_ds, rio.open(corr_filename, 'r') as fuse_ds:
        # test output image has been set with out_profile properties
        for k, v in out_profile.items():
            assert (
                (v is None and k not in fuse_ds.profile) or
                (fuse_ds.profile[k] == v) or
                (str(fuse_ds.profile[k]) == str(v))
            )  # yapf: disable

        # test output image has been set with src image properties not in out_profile
        if src_ds.profile['driver'].lower() == out_profile['driver'].lower():
            # source image keys including driver specific creation options, not present in out_profile
            src_keys = set(src_ds.profile.keys()).difference(out_profile.keys())
        else:
            # source image keys excluding driver specific creation options, not present in out_profile
            src_keys = {'width', 'height', 'count', 'dtype', 'crs', 'transform'}.difference(out_profile.keys())
        for k in src_keys:
            v = src_ds.profile[k]
            assert (
                (v is None and k not in fuse_ds.profile) or
                (fuse_ds.profile[k] == v) or
                (str(fuse_ds.profile[k]) == str(v))
            )  # yapf: disable


@pytest.mark.parametrize(
    'model, proc_crs', [
        (Model.gain, ProcCrs.ref),
        (Model.gain_blk_offset, ProcCrs.ref),
        (Model.gain_offset, ProcCrs.ref),
        (Model.gain, ProcCrs.src),
        (Model.gain_blk_offset, ProcCrs.src),
        (Model.gain_offset, ProcCrs.src),
    ]
)  # yapf: disable
def test_param_image(file_rgb_100cm_float, tmp_path: Path, model: Model, proc_crs: ProcCrs):
    """ Test creation and masking of parameter image for different model and proc_crs combinations. """
    corr_filename = tmp_path.joinpath('corrected.tif')
    param_filename = utils.create_param_filename(corr_filename)
    raster_fuse = RasterFuse(file_rgb_100cm_float, file_rgb_100cm_float, proc_crs=proc_crs)
    with raster_fuse:
        raster_fuse.process(corr_filename, model, (5, 5), param_filename=param_filename)

    assert (param_filename.exists())

    with rio.open(file_rgb_100cm_float, 'r') as ref_src_ds, rio.open(param_filename, 'r') as param_ds:
        assert (param_ds.count == ref_src_ds.count * 3)
        param_mask = param_ds.dataset_mask().astype('bool', copy=False)
        src_ref_mask = ref_src_ds.dataset_mask().astype('bool', copy=False)
        assert (param_mask == src_ref_mask).all()


@pytest.mark.parametrize(
    'src_file, ref_file, kernel_shape, proc_crs, mask_partial', [
        ('src_file_45cm_float', 'ref_file_100cm_float', (1, 1), ProcCrs.auto, False),
        ('src_file_45cm_float', 'ref_file_100cm_float', (1, 1), ProcCrs.auto, True),
        ('src_file_45cm_float', 'ref_file_100cm_float', (3, 3), ProcCrs.auto, True),
        ('src_file_100cm_float', 'ref_file_45cm_float', (1, 1), ProcCrs.auto, False),
        ('src_file_100cm_float', 'ref_file_45cm_float', (1, 1), ProcCrs.auto, True),
        ('src_file_100cm_float', 'ref_file_45cm_float', (3, 3), ProcCrs.auto, True),
    ]
)  # yapf: disable
def test_mask_partial(
    src_file: str, ref_file: str, tmp_path: Path, kernel_shape: Tuple[int, int], proc_crs: ProcCrs, mask_partial: bool,
    request: FixtureRequest,
):
    """ Test partial masking with multiple image blocks. """
    src_file: Path = request.getfixturevalue(src_file)
    ref_file: Path = request.getfixturevalue(ref_file)
    model_config = RasterFuse.create_model_config(mask_partial=mask_partial)
    block_config = RasterFuse.create_block_config(max_block_mem=1.e-1)
    corr_file = tmp_path.joinpath('corrected.tif')
    raster_fuse = RasterFuse(src_file, ref_file, proc_crs=proc_crs)
    with raster_fuse:
        raster_fuse.process(
            corr_file, Model.gain_blk_offset, kernel_shape, model_config=model_config, block_config=block_config
        )
    assert (corr_file.exists())
    with rio.open(src_file, 'r') as src_ds, rio.open(corr_file, 'r') as fuse_ds:
        fuse_mask = fuse_ds.dataset_mask().astype('bool', copy=False)
        src_mask = src_ds.dataset_mask().astype('bool', copy=False)
        if not mask_partial:
            assert (fuse_mask == src_mask).all()
        else:
            assert (fuse_mask.sum() < src_mask.sum())
            assert (fuse_mask.sum() > 0)
            assert (src_mask[fuse_mask]).all()
            # check that the output mask consists of 1 blob
            out_mask_shapes = [shape for shape in shapes(fuse_mask.astype('uint8', copy=False), mask=fuse_mask)]
            assert (len(out_mask_shapes) == 1)


def test_build_overviews(tmp_path: Path, ref_file_50cm_float):
    """ Test that overviews are built for corrected and parameter files. """
    corr_filename = tmp_path.joinpath('corrected.tif')
    param_filename = utils.create_param_filename(corr_filename)
    raster_fuse = RasterFuse(ref_file_50cm_float, ref_file_50cm_float)

    # replace raster_fuse.build_overviews() with a test_build_overviews() that forces min_level_pixels==1, otherwise
    # overviews won't be built for the small test raster
    orig_build_overviews = raster_fuse._build_overviews

    def test_build_overviews(im):
        orig_build_overviews(im, min_level_pixels=1)

    raster_fuse._build_overviews = test_build_overviews

    with raster_fuse:
        raster_fuse.process(corr_filename, Model.gain_blk_offset, (3, 3), param_filename=param_filename, build_ovw=True)

    assert (corr_filename.exists())
    assert (param_filename.exists())

    with rio.open(corr_filename, 'r') as fuse_ds:
        assert (len(fuse_ds.overviews(1)) > 0)

    with rio.open(param_filename, 'r') as param_ds:
        for band_i in param_ds.indexes:
            assert (len(param_ds.overviews(band_i)) > 0)


def test_io_error(tmp_path: Path, ref_file_50cm_float):
    """ Test we get an IoError if processing without opening/entering the context. """
    raster_fuse = RasterFuse(ref_file_50cm_float, ref_file_50cm_float)
    with pytest.raises(IoError):
        raster_fuse.process(tmp_path, Model.gain_blk_offset, (3, 3))


def test_corr_filename(tmp_path: Path, ref_file_50cm_float):
    """ Test corrected file is created. """
    corr_filename = tmp_path.joinpath('corrected.tif')
    raster_fuse = RasterFuse(ref_file_50cm_float, ref_file_50cm_float)
    with raster_fuse:
        raster_fuse.process(corr_filename, Model.gain_blk_offset, (3, 3))

    assert (corr_filename.exists())


def test_single_thread(tmp_path: Path, ref_file_50cm_float):
    """ Test single-threaded processing creates a corrected file. """
    block_config = RasterFuse.create_block_config(threads=1)
    corr_filename = tmp_path.joinpath('corrected.tif')
    raster_fuse = RasterFuse(ref_file_50cm_float, ref_file_50cm_float)
    with raster_fuse:
        raster_fuse.process(corr_filename, Model.gain_blk_offset, (3, 3), block_config=block_config)

    assert (corr_filename.exists())


@pytest.mark.parametrize(
    'src_file, ref_file, proc_crs, exp_proc_crs', [
        ('src_file_50cm_float', 'ref_file_100cm_float', ProcCrs.auto, ProcCrs.ref),
        ('src_file_50cm_float', 'ref_file_100cm_float', ProcCrs.src, ProcCrs.src),
        ('src_file_100cm_float', 'ref_file_50cm_float', ProcCrs.auto, ProcCrs.src),
        ('src_file_100cm_float', 'ref_file_50cm_float', ProcCrs.ref, ProcCrs.ref),
    ]
)  # yapf: disable
def test_proc_crs(
    tmp_path: Path, src_file: str, ref_file: str, proc_crs: ProcCrs, exp_proc_crs: ProcCrs, request: FixtureRequest,
):
    """ Test corrected file creation for forced and auto proc_crs with different src/ref combinations. """
    src_file: Path = request.getfixturevalue(src_file)
    ref_file: Path = request.getfixturevalue(ref_file)
    corr_filename = tmp_path.joinpath('corrected.tif')
    raster_fuse = RasterFuse(src_file, ref_file, proc_crs=proc_crs)
    assert (raster_fuse.proc_crs == exp_proc_crs)
    with raster_fuse:
        raster_fuse.process(corr_filename, Model.gain_blk_offset, (5, 5))
    assert (corr_filename.exists())


def test_tags(tmp_path: Path, ref_file_50cm_float):
    """ Test corrected file metadata. """
    model = Model.gain_blk_offset
    kernel_shape = (3, 3)
    proc_crs = ProcCrs.ref
    block_config = RasterFuse.create_block_config()
    raster_fuse = RasterFuse(ref_file_50cm_float, ref_file_50cm_float, proc_crs=proc_crs)
    corr_filename = tmp_path.joinpath('corrected.tif')
    param_filename = utils.create_param_filename(corr_filename)
    with raster_fuse:
        raster_fuse.process(
            corr_filename, model, kernel_shape, param_filename=param_filename, block_config=block_config
        )

    assert (corr_filename.exists())
    assert (param_filename.exists())
    utils.validate_param_image(param_filename)

    with rio.open(corr_filename, 'r') as out_ds:
        tags = out_ds.tags()
        assert (
            {
                'FUSE_SRC_FILE', 'FUSE_REF_FILE', 'FUSE_MODEL', 'FUSE_KERNEL_SHAPE', 'FUSE_PROC_CRS',
                'FUSE_MAX_BLOCK_MEM', 'FUSE_THREADS',
                *{f'FUSE_{k.upper()}' for k in RasterFuse.create_model_config().keys()},
            } <= set(tags)
        )
        assert (tags['FUSE_SRC_FILE'] == ref_file_50cm_float.name)
        assert (tags['FUSE_REF_FILE'] == ref_file_50cm_float.name)
        assert (tags['FUSE_MODEL'] == str(model.name))
        assert (tags['FUSE_PROC_CRS'] == str(proc_crs.name))
        assert (tags['FUSE_KERNEL_SHAPE'] == str(kernel_shape))
        for key, val in RasterFuse.create_model_config().items():
            assert (tags[f'FUSE_{key.upper()}'] == val.name if hasattr(val, 'name') else str(val))
        assert (yaml.safe_load(tags['FUSE_MAX_BLOCK_MEM']) == block_config['max_block_mem'])
        assert (yaml.safe_load(tags['FUSE_THREADS']) == block_config['threads'])

@pytest.mark.parametrize(
    'src_file, ref_file, src_bands, ref_bands, force, exp_bands', [
        ('file_rgb_50cm_float', 'file_rgb_100cm_float', None, None, False, (1, 2, 3)),
        ('file_rgb_50cm_float', 'file_rgb_100cm_float', (3, 2, 1), None, False, (3, 2, 1)),
        ('file_rgb_50cm_float', 'file_rgb_100cm_float', None, (3, 2, 1), False, (1, 2, 3)),
        ('file_rgb_50cm_float', 'file_rgb_100cm_float', (2, 1), (3, 1, 2), False, (2, 1)),
        ('file_rgb_50cm_float', 'file_rgb_100cm_float', (2, 1), (3, 2, 1), True, (3, 2)),
    ]
)  # yapf: disable
def test_src_ref_bands(
    src_file: str, ref_file: str, src_bands: Tuple[int], ref_bands: Tuple[int], force: bool, exp_bands: Tuple[int],
    tmp_path: Path, request: FixtureRequest
):
    """ Test fusion with the src_bands and ref_bands parameters. """
    src_file: Path = request.getfixturevalue(src_file)
    ref_file: Path = request.getfixturevalue(ref_file)
    corr_filename = tmp_path.joinpath('corrected.tif')
    with RasterFuse(src_file, ref_file, src_bands=src_bands, ref_bands=ref_bands, force=force) as raster_fuse:
        raster_fuse.process(corr_filename, model=Model.gain_blk_offset, kernel_shape=(3, 3))
    assert (corr_filename.exists())
    # open src_file in a WarpedVRT to reproject it North-up (if necessary)
    with WarpedVRT(rio.open(src_file, 'r')) as src_ds, rio.open(corr_filename, 'r') as out_ds:
        src_array = src_ds.read(indexes=exp_bands)
        src_mask = src_ds.dataset_mask().astype('bool', copy=False)
        out_array = out_ds.read()
        out_mask = out_ds.dataset_mask().astype('bool', copy=False)

        assert out_ds.count == len(exp_bands)
        assert (out_mask == src_mask).all()
        assert (out_array[:, out_mask] == pytest.approx(src_array[:, src_mask], abs=2))



"""
    Homonim: Correction of aerial and satellite imagery to surface reflectance
    Copyright (C) 2021 Dugal Harris
    Email: dugalh@gmail.com

    This program is free software: you can redistribute it and/or modify
    it under the terms of the GNU Affero General Public License as
    published by the Free Software Foundation, either version 3 of the
    License, or any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU Affero General Public License for more details.

    You should have received a copy of the GNU Affero General Public License
    along with this program.  If not, see <https://www.gnu.org/licenses/>.
"""

import concurrent.futures
import logging
import pathlib
import threading
from typing import Dict, List, Tuple

import numpy as np
from rasterio.warp import Resampling
from tabulate import tabulate
from tqdm.auto import tqdm

from homonim import utils
from homonim.enums import ProcCrs
from homonim.raster_pair import BlockPair
from homonim.matched_pair import MatchedPairReader

logger = logging.getLogger(__name__)


class RasterCompare(MatchedPairReader):

    def __init__(self, *args, **kwargs):
        """
        Class to compare source and reference images.

        Reference and source image(s) should be co-located and spectrally similar.  Reference image extents must
        encompass those of the source image(s).

        The reference image should contain bands that are approximate (wavelength) matches to the source image bands.
        Where source and reference images are RGB, or have ``center_wavelength`` metadata, bands are matched
        automatically based on wavelength.  Where there are the same number of source and reference bands, and no
        ``center_wavelength`` metadata, bands are assumed to be in matching order.  Subsets and ordering of source
        and reference bands can be specified with the ``src_bands`` and ``ref_bands`` parameters.

        .. note::

            Images downloaded with `geedim <https://github.com/leftfield-geospatial/geedim>`_ have ``center_wavelength`` metadata
            compatible with ``homonim``.

        Parameters
        ----------
        src_filename: str, pathlib.Path
            Path or URL of the source image file.
        ref_filename: str, pathlib.Path
            Path or URL of the reference image file.
        proc_crs: homonim.enums.ProcCrs, optional
            :class:`~homonim.enums.ProcCrs` instance specifying which of the source/reference image CRS and pixel
            grid to use for processing.  For most use cases, it can be left as the default of
            :attr:`~homonim.enums.ProcCrs.auto` i.e. the lowest resolution of the source and reference image CRS's.
        src_bands: list of int, optional.
            Indexes of source spectral bands to be processed (1 based).  If not specified, all bands with the
            ``center_wavelength`` property, or all non-alpha bands, are used.
        ref_bands: list of int, optional.
            Indexes of reference spectral bands to match and fuse with source bands (1 based).  Should contain at
            least as many elements as ``src_bands``, or the number of valid bands in the source image file,
            if ``src_bands`` is not specified.  If ``ref_bands`` is not specified, all reference bands with the
            ``center_wavelength`` property, or all non-alpha bands are used.
        force: bool, optional
            Bypass auto wavelength matching, and any band-matching errors.  Use with caution.
        """
        MatchedPairReader.__init__(self, *args, **kwargs)
        self._lock = threading.Lock()

    schema = dict(
        r2=dict(abbrev='r\N{SUPERSCRIPT TWO}', description='Pearson\'s correlation coefficient squared', ),
        rmse=dict(abbrev='RMSE', description='Root Mean Square Error', ),
        rrmse=dict(abbrev='rRMSE', description='Relative RMSE (RMSE/mean(ref))', ),
        n=dict(abbrev='N', description='Number of pixels', )
    )  # yapf: disable
    """ Dictionary describing the statistics returned by :meth:`RasterCompare.process`. """

    @staticmethod
    def schema_table() -> str:
        """ Return a table string describing the :meth:`RasterCompare.process` statistics. """
        headers = {key: key.upper() for key in list(RasterCompare.schema.values())[0].keys()}
        return tabulate(RasterCompare.schema.values(), headers=headers, tablefmt=utils.table_format)

    @staticmethod
    def create_config(
        threads: int = 0, max_block_mem: float = 512, downsampling: Resampling = Resampling.average,
        upsampling: Resampling = Resampling.cubic_spline,
    ) -> Dict:
        """
        Utility method to create a RasterCompare configuration dictionary whose items can be passed as keyword
        arguments to :meth:`RasterCompare.process`.  Without arguments, the default configuration is returned.

        Parameters
        ----------
        threads: int, optional
            Number of image blocks to process concurrently.  A maximum of the number of processors on your
            system is allowed.  Increasing this number will increase the memory required for processing.
            0 = use all processors.
        max_block_mem: float, optional
            Maximum size of an image block in megabytes. Note that the total memory consumed by a thread is
            proportional to, but larger than this number.
        downsampling: rasterio.enums.Resampling, optional
            Resampling method to use when downsampling. See the `rasterio docs
            <https://rasterio.readthedocs.io/en/latest/api/rasterio.enums.html#rasterio.enums.Resampling>`_ for
            available options.
        upsampling: rasterio.enums.Resampling, optional
            Resampling method to use when upsampling.  See the `rasterio docs
            <https://rasterio.readthedocs.io/en/latest/api/rasterio.enums.html#rasterio.enums.Resampling>`_ for
            available options.

        Returns
        -------
        dict
            Configuration dictionary.
        """
        return dict(
            threads=utils.validate_threads(threads), max_block_mem=max_block_mem, downsampling=downsampling,
            upsampling=upsampling,
        )

    def _get_resampling(self, from_res: Tuple[float, float], to_res: Tuple[float, float], **kwargs) -> Resampling:
        """
        Utility method to return the resampling method for re-projecting from resolution `from_res` to resolution
        `to_res`.
        """
        config = self.create_config(**kwargs)
        return config['downsampling'] if np.prod(np.abs(from_res)) <= np.prod(np.abs(to_res)) else config['upsampling']

    def _get_image_stats(self, image_sums: List[Dict]) -> List[Dict]:
        """ Return the band comparison statistics, given src, ref, src**2 etc band sums. """

        def get_band_stats(
            src_sum: float = 0, ref_sum: float = 0, src2_sum: float = 0, ref2_sum: float = 0, src_ref_sum: float = 0,
            res2_sum: float = 0, mask_sum: float = 0
        ) -> Dict:
            """ Return the comparison statistics for a band, given the source, reference etc. sums. """
            # find PCC using the 3rd equation down at
            # https://en.wikipedia.org/wiki/Pearson_correlation_coefficient#For_a_sample
            src_mean = src_sum / mask_sum
            ref_mean = ref_sum / mask_sum
            pcc_num = src_ref_sum - (mask_sum * src_mean * ref_mean)
            pcc_den = (
                np.sqrt(src2_sum - (mask_sum * (src_mean ** 2))) * np.sqrt(ref2_sum - (mask_sum * (ref_mean ** 2)))
            )
            pcc = pcc_num / pcc_den

            # find RMSE and rRMSE
            rmse = np.sqrt(res2_sum / mask_sum)
            rrmse = rmse / ref_mean
            return dict(r2=pcc ** 2, rmse=rmse, rrmse=rrmse, n=int(mask_sum))

        image_stats = {}
        sum_over_bands = {}
        for band_i, band_sum_dict in enumerate(image_sums):
            band_stats = get_band_stats(**band_sum_dict)
            # prefer ref band name as key: with multiple source images compared with one reference, this makes the
            # comparison tables easier to interpret
            band_desc = (
                self.ref_im.descriptions[self.ref_bands[band_i] - 1] or
                self.src_im.descriptions[self.src_bands[band_i] - 1] or
                f'Ref. band {self.ref_bands[band_i]}'
            )  # yapf: disable
            image_stats[band_desc] = band_stats
            sum_over_bands = {k: sum_over_bands.get(k, 0) + v for k, v in band_stats.items()}

        # find mean of each statistic over the bands, retaining int types
        mean_stats = {
            k: int(v / len(image_sums)) if isinstance(v, int) else (v / len(image_sums))
            for k, v in sum_over_bands.items()
        }  # yapf: disable
        # add the means to the list of bands
        image_stats['Mean'] = mean_stats
        return image_stats

    @staticmethod
    def stats_table(stats_dict: Dict[str, Dict], key_header: str = 'band'):
        """
        Return a table string for the provided comparison statistics.

        Parameters
        ----------
        stats_dict: dict of str: dict
            Comparison statistics to tabulate, as returned by :meth:`RasterCompare.process`.
        key_header: str, optional
            Table header for the stats_dict key values.

        Returns
        -------
        str
            Table string.
        """
        stats_list = [dict(**{key_header: key}, **val) for key, val in stats_dict.items()]
        headers = {
            k: RasterCompare.schema[k]['abbrev'] if k in RasterCompare.schema else str.capitalize(k)
            for k in list(stats_list[0].keys())
        }  # yapf: disable
        return tabulate(stats_list, headers=headers, floatfmt='.3f', stralign='right', tablefmt=utils.table_format)

    def process(self, **kwargs) -> Dict[str, Dict]:
        """
        Compare source and reference images.

        To improve speed and reduce memory usage, images are divided into blocks for concurrent processing.

        Parameters
        ----------
        kwargs
            Optional configuration settings.  See :meth:`RasterCompare.create_config` for possible arguments and their
            default values.

        Returns
        -------
        dict of dict
            Dict representing the comparison statistics.
        """
        self._assert_open()
        config = self.create_config(**kwargs)

        def get_block_sums(block_pair: BlockPair):
            """ Thread-safe function to find the source/reference sums for a block.  """
            src_ra, ref_ra = self.read(block_pair)  # read src and ref blocks

            # re-project so that both source and reference are in proc_crs
            if self.proc_crs == ProcCrs.ref:
                resampling = self._get_resampling(src_ra.res, ref_ra.res, **kwargs)
                src_ra = src_ra.reproject(**ref_ra.proj_profile, resampling=resampling)
            else:
                resampling = self._get_resampling(ref_ra.res, src_ra.res, **kwargs)
                ref_ra = ref_ra.reproject(**src_ra.proj_profile, resampling=resampling)

            # mask invalid pixels so they don't contribute to sums
            src_array = src_ra.array
            ref_array = ref_ra.array
            mask = ref_ra.mask & src_ra.mask
            src_array[~mask] = 0
            ref_array[~mask] = 0
            # find the required sums and return
            sums_dict = dict(
                src_sum=src_array.sum(), ref_sum=ref_array.sum(), src2_sum=(src_array ** 2).sum(),
                ref2_sum=(ref_array ** 2).sum(), src_ref_sum=(src_array * ref_array).sum(),
                res2_sum=((ref_array - src_array) ** 2).sum(), mask_sum=mask.sum()
            )
            return sums_dict, block_pair

        with concurrent.futures.ThreadPoolExecutor(max_workers=config['threads']) as executor:
            # read and sum image blocks in threads
            futures = [
                executor.submit(get_block_sums, block_pair)
                for block_pair in self.block_pairs(max_block_mem=config['max_block_mem'])
            ]  # yapf: disable

            # wait for threads
            image_sums = [{} for _ in self.src_bands]
            bar_format = '{l_bar}{bar}|{n_fmt}/{total_fmt} blocks [{elapsed}<{remaining}]'
            for future in tqdm(
                concurrent.futures.as_completed(futures), bar_format=bar_format, total=len(futures), dynamic_ncols=True,
            ):  # yapf: disable
                # get block sums and accumulate over the image
                block_sums_dict, block_pair = future.result()
                image_sums[block_pair.band_i] = (
                    {k: image_sums[block_pair.band_i].get(k, 0) + v for k, v in block_sums_dict.items()}
                )  # yapf: disable

        # find the comparison statistics from the accumulated block sums, and return
        return self._get_image_stats(image_sums)



"""
    Homonim: Correction of aerial and satellite imagery to surface reflectance
    Copyright (C) 2021 Dugal Harris
    Email: dugalh@gmail.com

    This program is free software: you can redistribute it and/or modify
    it under the terms of the GNU Affero General Public License as
    published by the Free Software Foundation, either version 3 of the
    License, or any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU Affero General Public License for more details.

    You should have received a copy of the GNU Affero General Public License
    along with this program.  If not, see <https://www.gnu.org/licenses/>.
"""

import json
from pathlib import Path
from typing import Dict, List

import pytest
import rasterio as rio
from click.testing import CliRunner
from pytest import FixtureRequest
from rasterio.windows import get_data_window, Window

from homonim.cli import cli
from homonim.errors import ImageFormatError, IoError
from homonim.stats import ParamStats
from tests.conftest import str_contain_no_space


def _test_vals(param_stats: List[Dict]):
    """ Helper function to test statistics against known values for param_file (i.e. gain=1, offset=0, r2=1). """
    assert len(param_stats) == 9
    for band_stats in param_stats:
        assert ({'band', 'mean', 'std', 'min', 'max'} <= set(band_stats.keys()))

    exp_param_stats_list = (
        3 * [{'mean': 1, 'std': 0, 'min': 1, 'max': 1}] +  # gains
        3 * [{'mean': 0, 'std': 0, 'min': 0, 'max': 0}] +  # offsets
        3 * [{'mean': 1, 'std': 0, 'min': 1, 'max': 1, 'inpaint_p': 0}]  # r2
    )  # yapf: disable

    for param_band_stats, exp_param_band_stats in zip(param_stats, exp_param_stats_list):
        for k, v in exp_param_band_stats.items():
            assert (param_band_stats[k] == pytest.approx(exp_param_band_stats[k], abs=1e-2))


@pytest.mark.parametrize('param_file_str', ['param_file', 'param_file_tile_10x20'])
def test_api__stats(param_file_str: str, request: FixtureRequest):
    """ Test ParamStats creation and execution. """
    param_file: Path = request.getfixturevalue(param_file_str)
    with ParamStats(param_file) as stats:
        assert (len(stats.metadata) > 0)
        param_stats = stats.stats()
    _test_vals(param_stats)


def test_api__context(param_file: Path):
    """ Test ParamStats context management. """
    stats = ParamStats(param_file)
    # test ParamStats usage without context entry raises an IoError
    with pytest.raises(IoError):
        _ = stats.stats()
    # test ParamStats.stats usage with context entry is ok, and file is closed on context exit
    with stats:
        assert (len(stats.metadata) > 0)
        _ = stats.stats()
    assert stats.closed


def test_api__tables(param_file: Path):
    """ Test ParamStats table generation. """
    with ParamStats(param_file) as stats:
        assert (len(stats.metadata) > 0)
        param_stats = stats.stats()

    schema_table = stats.schema_table()
    assert len(schema_table) > 0

    param_table = stats.stats_table(param_stats)
    assert len(param_table) > 0

    # test stats data are in table
    for band_stats in param_stats:
        for k, v in band_stats.items():
            assert (f'{v:.3f}' in param_table) if isinstance(v, float) else (v in param_table)

    # test stats abbreviations are in table
    for schema_dict in stats.schema.values():
        assert schema_dict['abbrev'] in param_table


@pytest.mark.parametrize('threads', [1, 0])
def test_api__threads(param_file: Path, threads: int):
    """ Test ParamStats.stats works multi- and single-threaded. """
    with ParamStats(param_file) as stats:
        assert (len(stats.metadata) > 0)
        param_stats = stats.stats(threads=threads)
    _test_vals(param_stats)


@pytest.mark.parametrize('param_file_str', ['param_file', 'param_file_tile_10x20'])
def test_api__data_window(param_file_str: str, request: FixtureRequest):
    """ Test ParamStats._get_data_window() accumulates block windows correctly. """
    param_file: Path = request.getfixturevalue(param_file_str)
    with rio.open(param_file, 'r') as param_im:
        mask = param_im.read_masks(indexes=1)
        data_win = get_data_window(mask, nodata=0)
        image_win = Window(0, 0, param_im.width, param_im.height)
        assert data_win != image_win

    with ParamStats(param_file) as stats:
        assert data_win == stats._get_data_window()


def test_api__file_format_error(file_rgb_100cm_float):
    """ Test incorrect parameter file format raises an error. """
    with pytest.raises(ImageFormatError):
        _ = ParamStats(file_rgb_100cm_float)


@pytest.mark.parametrize('param_file_str', ['param_file', 'param_file_tile_10x20'])
def test_cli(runner: CliRunner, param_file_str: str, request: FixtureRequest):
    """ Test stats cli generates the correct output. """
    param_file: Path = request.getfixturevalue(param_file_str)
    cli_str = f'stats {param_file}'
    result = runner.invoke(cli, cli_str.split())
    assert (result.exit_code == 0)
    assert (param_file.name in result.output)

    res_str = """B1_GAIN    1.000 0.000  1.000 1.000          
    B2_GAIN    1.000 0.000  1.000 1.000          
    B3_GAIN    1.000 0.000  1.000 1.000          
    B1_OFFSET -0.000 0.000 -0.001 0.000          
    B2_OFFSET -0.000 0.000 -0.001 0.000          
    B3_OFFSET -0.000 0.000 -0.001 0.000          
    B1_R2      1.000 0.000  1.000 1.000     0.000    
    B2_R2      1.000 0.000  1.000 1.000     0.000    
    B3_R2      1.000 0.000  1.000 1.000     0.000"""
    assert (str_contain_no_space(res_str, result.output))


def test_cli__out_file(tmp_path: Path, runner: CliRunner, param_file: Path):
    """ Test stats cli --output option generates the correct values. """
    output_file = tmp_path.joinpath('stats.json')
    cli_str = f'stats {param_file} --output {output_file}'
    result = runner.invoke(cli, cli_str.split())
    assert (result.exit_code == 0)
    assert (output_file.exists())

    with open(output_file) as f:
        stats_dict = json.load(f)

    param_file = str(param_file)
    assert (param_file in stats_dict)
    param_stats = stats_dict[param_file]
    _test_vals(param_stats)


def test_cli__mult_inputs(tmp_path: Path, runner: CliRunner, param_file: Path):
    """ Test stats cli with multiple input files. """
    output_file = tmp_path.joinpath('stats.json')
    cli_str = f'stats {param_file} {param_file} --output {output_file}'
    result = runner.invoke(cli, cli_str.split())
    assert (result.exit_code == 0)
    assert (output_file.exists())

    with open(output_file) as f:
        stats_dict = json.load(f)

    param_file = str(param_file)
    assert (param_file in stats_dict)


def test_cli__file_format_error(runner: CliRunner, file_rgb_100cm_float):
    """ Test stats cli fails with error message when the parameter file format is incorrect. """
    cli_str = f'stats {file_rgb_100cm_float}'
    result = runner.invoke(cli, cli_str.split())
    assert (result.exit_code != 0)
    assert ('Invalid value' in result.output)


##



"""
    Homonim: Correction of aerial and satellite imagery to surface reflectance
    Copyright (C) 2021 Dugal Harris
    Email: dugalh@gmail.com

    This program is free software: you can redistribute it and/or modify
    it under the terms of the GNU Affero General Public License as
    published by the Free Software Foundation, either version 3 of the
    License, or any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU Affero General Public License for more details.

    You should have received a copy of the GNU Affero General Public License
    along with this program.  If not, see <https://www.gnu.org/licenses/>.
"""

import json
import logging
import pathlib
import re
import sys
from timeit import default_timer as timer
from typing import Tuple, Dict

import click
import cloup
import numpy as np
import rasterio as rio
import yaml
from click.core import ParameterSource
from rasterio.warp import SUPPORTED_RESAMPLING

from homonim import utils, version, RasterFuse, RasterCompare, ParamStats, ProcCrs, Model
from homonim.errors import ImageFormatError
from homonim.raster_array import RasterArray
from homonim.kernel_model import KernelModel

logger = logging.getLogger(__name__)


class PlainInfoFormatter(logging.Formatter):
    """ logging formatter to format INFO logs without the module name etc. prefix. """

    def format(self, record: logging.LogRecord):
        if record.levelno == logging.INFO:
            self._style._fmt = '%(message)s'
        else:
            self._style._fmt = '%(levelname)s:%(name)s: %(message)s'
        return super().format(record)


class HomonimCommand(cloup.Command):
    """ cloup.Command subclass for formatting help with RST markup. """

    def get_help(self, ctx: click.Context):
        """ Strip some RST markup from the help text for CLI display.  Will not work with grid tables. """

        # Note that this can't easily be done in __init__, as each sub-command's __init__ gets called,
        # which ends up re-assigning self.wrap_text to reformat_text
        if not hasattr(self, 'wrap_text'):
            self.wrap_text = cloup.formatting._formatter.wrap_text
        sub_strings = {
            '\b\n': '\n\b',                 # convert from RST friendly to click literal (unwrapped) block marker
            r'\| ': '',                     # strip RST literal (unwrapped) marker in e.g. tables and bullet lists
            '\n\.\. _.*:\n': '',            # strip RST ref directive '\n.. _<name>:\n'
            '`(.*?) <(.*?)>`_': r'\g<1>',   # convert from RST cross-ref '`<name> <<link>>`_' to 'name'
            '::': ':',                      # convert from RST '::' to ':'
            '``(.*?)``': r'\g<1>',          # convert from RST '``literal``' to 'literal'
            ':option:`(.*?)( <.*?>)?`': r'\g<1>',   # convert ':option:`--name <group-command --name>`' to '--name'
            ':option:`(.*?)`': r'\g<1>',    # convert ':option:`--name`' to '--name'
            '-{4,}': r'',                   # strip '----...'
        }  # yapf: disable

        def reformat_text(text: str, width: int, **kwargs):
            for sub_key, sub_value in sub_strings.items():
                text = re.sub(sub_key, sub_value, text, flags=re.DOTALL)
            wr_text = self.wrap_text(text, width, **kwargs)
            # change double newline to single newline separated list
            return re.sub('\n\n(\s*?)- ', '\n- ', wr_text, flags=re.DOTALL)

        cloup.formatting._formatter.wrap_text = reformat_text
        return cloup.Command.get_help(self, ctx)


class FuseCommand(HomonimCommand):
    """ click.Command subclass for setting fuse command parameters from a yaml config file. """

    def invoke(self, ctx: click.Context):
        """ Merge config file with command line and default parameter values.  """
        # adapted from https://stackoverflow.com/questions/46358797/python-click-supply-arguments-and-options-from-a
        # -configuration-file/46391887
        config_file = ctx.params['conf']
        if config_file is not None:
            # read the config file into a dict
            with open(config_file) as f:
                config_dict = yaml.safe_load(f)

            for conf_key, conf_value in config_dict.items():
                if conf_key not in ctx.params:
                    raise click.BadParameter(f'Unknown config file parameter "{conf_key}"', ctx=ctx, param_hint='conf')
                else:
                    param_src = ctx.get_parameter_source(conf_key)
                    # overwrite parameters not specified on command line with config file values
                    if ctx.params[conf_key] is None or param_src == ParameterSource.DEFAULT:
                        ctx.params[conf_key] = conf_value
                        ctx.set_parameter_source(conf_key, ParameterSource.COMMANDLINE)

        # Set the default creation_options if no other driver or creation_options have been specified.
        # This can't be done in a callback as it depends on --driver.
        if (ctx.get_parameter_source('driver') == ParameterSource.DEFAULT and
                ctx.get_parameter_source('creation_options') == ParameterSource.DEFAULT):
            ctx.params['creation_options'] = RasterFuse.create_out_profile()['creation_options']

        return click.Command.invoke(self, ctx)


def _update_existing_keys(default_dict: Dict, **kwargs) -> Dict:
    """ Update values in a dict with args from matching keys in **kwargs. """
    return {k: kwargs.get(k, v) for k, v in default_dict.items()}


def _configure_logging(verbosity: int):
    """ Configure python logging level."""
    # adapted from rasterio https://github.com/rasterio/rasterio
    log_level = max(10, 20 - 10 * verbosity)

    # apply config to package logger, rather than root logger
    pkg_logger = logging.getLogger(__package__)
    formatter = PlainInfoFormatter()
    handler = logging.StreamHandler(sys.stderr)
    handler.setFormatter(formatter)
    pkg_logger.addHandler(handler)
    pkg_logger.setLevel(log_level)
    logging.captureWarnings(True)


def _threads_cb(ctx: click.Context, param: click.Option, value):
    """ click callback to validate --threads. """
    try:
        threads = utils.validate_threads(value)
    except Exception as ex:
        raise click.BadParameter(str(ex))
    return threads


def _nodata_cb(ctx: click.Context, param: click.Option, value):
    """ click callback to convert --nodata value to None, nan or float. """
    # adapted from rasterio https://github.com/rasterio/rasterio
    if value is None or value.lower() in ['null', 'nil', 'none', 'nada']:
        return None
    else:
        # check value is a number and can be cast to output dtype
        try:
            value = float(value.lower())
            if not rio.dtypes.can_cast_dtype(value, ctx.params['dtype']):
                raise click.BadParameter(
                    f'{value} cannot be cast to the output image data type {ctx.params["dtype"]}', param=param,
                    param_hint='nodata'
                )
        except (TypeError, ValueError):
            raise click.BadParameter(f'{value} is not a number', param=param, param_hint='nodata')

        return value


def _creation_options_cb(ctx: click.Context, param: click.Option, value):
    """
    click callback to validate and parse multiple creation options (e.g. `-co KEY1=VAL1 -co KEY2=VAL2).
    Note: `==VAL` breaks this as `str.split('=', 1)` is used.
    """
    # adapted from rasterio https://github.com/rasterio/rasterio
    if not value:
        return {}
    else:
        out = {}
        for pair in value:
            if '=' not in pair:
                raise click.BadParameter('Invalid syntax for KEY=VAL arg: {}'.format(pair))
            else:
                k, v = pair.split('=', 1)
                k = k.lower()
                v = v.lower()
                out[k] = None if v.lower() in ['none', 'null', 'nil', 'nada'] else yaml.safe_load(v)
        return out


def _param_file_cb(ctx: click.Context, param: click.Argument, value):
    """ click callback to validate parameter image file(s). """
    for filename in value:
        filename = pathlib.Path(filename)
        try:
            utils.validate_param_image(filename)
        except (FileNotFoundError, ImageFormatError):
            raise click.BadParameter(f'{filename.name} is not a valid parameter image.', param=param)
    return value

# define click options and arguments common to more than one command
# use cloup's argument to auto print argument help on command line
ref_file_arg = cloup.argument(
    'ref-file', nargs=1, metavar='REFERENCE', type=click.Path(exists=False, dir_okay=False, path_type=pathlib.Path),
    help='Path or URL of a reference image.'
)
threads_option = click.option(
    '-t', '--threads', type=click.INT, default=RasterFuse.create_block_config()['threads'], show_default=True,
    callback=_threads_cb, help=f'Number of image blocks to process concurrently (0 = use all processors).'
)
max_block_mem_option = click.option(
    '-mbm', '--max-block-mem', type=click.FLOAT, default=RasterFuse.create_block_config()['max_block_mem'],
    show_default=True, help='Maximum image block size in megabytes (0 = block corresponds to a whole band).'
)
downsampling_option = click.option(
    '-ds', '--downsampling', type=click.Choice([r.name for r in rio.warp.SUPPORTED_RESAMPLING]),
    default=RasterFuse.create_model_config()['downsampling'].name, show_default=True,
    help='Resampling method for re-projecting from high to low resolution.  See the `rasterio docs '
    '<https://rasterio.readthedocs.io/en/latest/api/rasterio.enums.html#rasterio.enums.Resampling>`_ for '
    'details.'
)
upsampling_option = click.option(
    '-us', '--upsampling', type=click.Choice([r.name for r in rio.warp.SUPPORTED_RESAMPLING]),
    default=RasterFuse.create_model_config()['upsampling'].name, show_default=True,
    help='Resampling method for re-projecting from low to high resolution.  See the `rasterio docs '
    '<https://rasterio.readthedocs.io/en/latest/api/rasterio.enums.html#rasterio.enums.Resampling>`_ for '
    'details.'
)
output_option = click.option(
    '-op', '--output', type=click.Path(exists=False, dir_okay=False, writable=True, path_type=pathlib.Path),
    help='Write results to this json file.'
)
src_bands_option = click.option(
    '-sb', '--src-band', 'src_bands', type=click.INT, multiple=True, default=None,
    show_default='all spectral or non-alpha bands.', help=f'Source band index(es) to process (1 based).'
)
ref_bands_option = click.option(
    '-rb', '--ref-band', 'ref_bands', type=click.INT, multiple=True, default=None,
    show_default='all spectral or non-alpha bands.',
    help=f'Reference band index(es) to match with source band(s) (1 based).'
)
force_match_option = click.option(
    '-f', '--force-match',  is_flag=True, type=click.BOOL, default=False, show_default=True,
    help=f'Bypass auto wavelength matching, and any band-matching errors.  Use with caution.'
)
""" cloup context settings to print help in 'linear' layout with heading/option emphasis. """
context_settings = cloup.Context.settings(
    formatter_settings=cloup.HelpFormatter.settings(
        col2_min_width=np.inf,
        theme=cloup.HelpTheme(
            invoked_command=cloup.Style(fg='bright_white', bold=True),
            heading=cloup.Style(fg='bright_white', bold=True),
            col1=cloup.Style(fg='bright_white'),
        )
    )
)  # yapf: disable


# define the click CLI
@cloup.group(context_settings=context_settings)
@click.option('--verbose', '-v', count=True, help='Increase verbosity.')
@click.option('--quiet', '-q', count=True, help='Decrease verbosity.')
@click.version_option(version=version.__version__, message='%(version)s')
def cli(verbose: int, quiet: int):
    """ Surface reflectance correction toolkit. """
    verbosity = verbose - quiet
    _configure_logging(verbosity)


# fuse command
@cloup.command(cls=FuseCommand)
# standard options
@cloup.argument(
    'src-file', nargs=-1, metavar='SOURCE...', type=click.Path(exists=False, dir_okay=False, path_type=pathlib.Path),
    help='Path/URL(s) of source image(s) to be corrected.'
)
@ref_file_arg
@cloup.option_group(
    "Standard options",
    # note: either use click.option(...), or cloup.option(..., help=inspect.cleandoc(...)) for RST help strings,
    # if cloup's mutually exclusive etc. functionality is needed, it should be the latter.
    click.option(
        '-m', '--model', type=click.Choice([m.value for m in Model], case_sensitive=False),
        default=KernelModel.default_model.value, show_default=True, help="""Correction model.

        - `gain`: Gain-only model, suitable for haze-free and zero offset images.

        - `gain-blk-offset`: Gain-only model applied to offset normalised blocks.  Suitable for most source-reference combinations.

        - `gain-offset`: Gain and offset model.  Most accurate model, but sensitive to differences between source and reference.
        """,
    ),  # yapf: disable
    click.option(
        '-k', '--kernel-shape', type=click.Tuple([click.INT, click.INT]), nargs=2,
        default=KernelModel.default_kernel_shape, show_default=True, metavar='HEIGHT WIDTH',
        help='Kernel height and width in pixels of the :option:`--proc-crs <homonim-fuse --proc-crs>` image. Larger '
             'kernels are less susceptible to over-fitting, but provide lower resolution correction.'
    ),
    src_bands_option,
    ref_bands_option,
    click.option(
        '-od', '--out-dir', type=click.Path(exists=True, file_okay=False, writable=True),
        show_default='source image directory.', help='Directory in which to place corrected image(s).'
    ),
    click.option(
        '-o', '--overwrite', is_flag=True, type=click.BOOL, default=False, show_default=True,
        help='Overwrite existing output file(s).'
    ),
    click.option(
        '-cmp', '--compare', 'cmp_file', metavar='FILE',
        type=click.Path(exists=False, dir_okay=False, path_type=pathlib.Path), is_flag=False, flag_value='ref',
        default=None,
        help='Compare source and corrected images with this reference image.  If no ``FILE`` value is given, source '
        'and corrected images are compared with :option:`REFERENCE`.'
    ),
    click.option(
        '-cb', '--cmp-band', 'cmp_bands', type=click.INT, multiple=True, default=None,
        show_default='all spectral or non-alpha bands.',
        help=f'Comparison reference band index(es) that correspond (spectrally) to '
             f':option:`--src-band <homonim-fuse --src-band>` (s).'
    ),
    click.option(
        '-bo/-nbo', '--build-ovw/--no-build-ovw', type=click.BOOL, default=True, show_default=True,
        help='Build overviews for the output image(s).'
    ),
    click.option(
        '-c', '--conf', type=click.Path(exists=True, dir_okay=False, readable=True, path_type=pathlib.Path),
        required=False, default=None, show_default=True,
        help='Path to a yaml configuration file specifying advanced options (as follow below).'
    )
)
# advanced options
@cloup.option_group(
    "Advanced options",
    click.option(
        '-pi/-npi', '--param-image/--no-param-image', type=click.BOOL, default=False, show_default=True,
        help=f'Write the  model parameters and R\N{SUPERSCRIPT TWO} values for each corrected image to a parameter '
        f'image file.'
    ),
    click.option(
        '-mp/-nmp', '--mask-partial/--no-mask-partial', type=click.BOOL,
        default=RasterFuse.create_model_config()['mask_partial'], show_default=True,
        help=f'Mask output pixels produced from partial kernel or source / reference image coverage.'
    ),
    # yapf: disable
    threads_option, max_block_mem_option, downsampling_option, upsampling_option,
    # yapf: enable
    click.option(
        '-rit', '--r2-inpaint-thresh', type=click.FloatRange(min=0, max=1),
        default=RasterFuse.create_model_config()['r2_inpaint_thresh'], show_default=True, metavar='FLOAT 0-1',
        help='R\N{SUPERSCRIPT TWO} threshold below which to inpaint model parameters from surrounding areas '
        '(0 = turn off inpainting). Valid for `gain-offset` :option:`--model` only.'
    ),
    click.option(
        '-pc', '--proc-crs', type=click.Choice([pc.value for pc in ProcCrs], case_sensitive=False),
        default=ProcCrs.auto.value, show_default=True,
        help="""The image CRS in which to estimate correction parameters. 
        \b
    
        - `auto`: lowest resolution of the source and reference CRS's (recommended). 
        - `src`: source image CRS. 
        - `ref`: reference image CRS.
        """
    ),
    click.option(
        '--driver',
        type=click.Choice(tuple(set(rio.drivers.raster_driver_extensions().values())), case_sensitive=False),
        default=RasterFuse.create_out_profile()['driver'], show_default=True, metavar='TEXT',
        help='Output image format driver.  See the `GDAL docs <https://gdal.org/drivers/raster/index.html>`_ for '
        'details.'
    ),  # yapf: disable
    click.option(
        '--dtype', type=click.Choice(list(rio.dtypes.dtype_fwd.values())[1:8], case_sensitive=False),
        default=RasterFuse.create_out_profile()['dtype'], show_default=True,
        help=f'Output image data type.  Valid for corrected images only, parameter images always use '
        f'{RasterArray.default_dtype}.'
    ),
    click.option(
        '--nodata', 'nodata', type=click.STRING, callback=_nodata_cb, metavar='[NUMBER|null|nan]',
        default=RasterFuse.create_out_profile()['nodata'], show_default=True,
        help=f'Output image nodata value.  Valid for corrected images only, parameter images always use '
        f'{RasterArray.default_nodata}.'
    ),
    click.option(
        '-co', '--creation-options', metavar='NAME=VALUE', multiple=True, default=(), callback=_creation_options_cb,
        help='Driver specific image creation option(s) for the output image(s).  See the `GDAL docs '
        '<https://gdal.org/drivers/raster/index.html>`_ for details.'
    ),
    force_match_option,  # yapf:disable
)
@click.pass_context
def fuse(
    ctx: click.Context, src_file: Tuple[pathlib.Path, ...], ref_file: pathlib.Path, model: Model,
    kernel_shape: Tuple[int, int], src_bands: Tuple[int], ref_bands: Tuple[int], out_dir: pathlib.Path,
    overwrite: bool, cmp_file: pathlib.Path, cmp_bands: Tuple[int], build_ovw: bool, proc_crs: ProcCrs,
    conf: pathlib.Path, param_image: bool, force_match: bool, **kwargs
):
    # @formatter:off
    """
    Correct image(s) to surface reflectance.

    Correct source multi-spectral aerial or satellite imagery to approximate surface reflectance, by fusion with a
    reference satellite image.

    For best results, reference and source image(s) should be concurrent, co-located and spectrally similar.  Reference
    image extents must encompass those of the source image(s).

    The reference image should contain bands that are approximate (wavelength) matches to the source image bands.
    Where source and reference images are RGB, or have ``center_wavelength`` metadata, bands are matched automatically.
    Where there are the same number of source and reference bands, and no ``center_wavelength`` metadata, bands are
    assumed to be in matching order.  The :option:`--src-band <homonim-fuse --src-band>` and
    :option:`--ref-band <homonim-fuse --ref-band>` options allow subsets and ordering of source and reference bands
    to be specified.

    Corrected image(s) are named automatically based on the source file name and option values.
    \b

    Examples:
    ---------

    Correct `source.tif` with `reference.tif` using the default options::

        homonim fuse source.tif reference.tif

    Correct `source.tif` with `reference.tif` using the `gain-blk-offset` model, a kernel of 5 x 5 pixels,
    and place the corrected images in the `corrected` directory::

        homonim fuse --model gain-blk-offset --kernel-shape 5 5 --out-dir ./corrected source.tif reference.tif

    Correct files matching `source*.tif` with `reference1.tif` using the `gain-offset` model and a kernel of 15 x 15
    pixels.  Produce parameter images, mask partially covered pixels in the corrected images, and statistically
    compare source and corrected images with `reference2.tif`::

        homonim fuse -m gain-offset -k 15 15 --param-image --mask-partial --compare reference2.tif source*.tif reference1.tif

    Correct bands 2 and 3 of `source.tif`, with bands 7 and 8 of `reference.tif`, using the default correction options::

        homonim fuse -sb 2 -sb 3 -rb 7 -rb 8 source.tif reference.tif

    ----
    """
    # @formatter:on

    # try:
    #     kernel_shape = utils.validate_kernel_shape(kernel_shape, model=model)
    # except Exception as ex:
    #     raise click.BadParameter(str(ex))

    # build configuration dictionaries for RasterFuse
    block_config = _update_existing_keys(RasterFuse.create_block_config(), **kwargs)
    model_config = _update_existing_keys(RasterFuse.create_model_config(), **kwargs)
    out_profile = _update_existing_keys(RasterFuse.create_out_profile(), **kwargs)
    config = dict(block_config=block_config, model_config=model_config, out_profile=out_profile)
    comp_files = []

    # iterate over and correct source file(s)
    try:
        for src_filename in src_file:
            out_path = pathlib.Path(out_dir) if out_dir is not None else src_filename.parent
            logger.info(f'\nCorrecting {src_filename.name}')
            with RasterFuse(
                src_filename, ref_file, proc_crs=proc_crs, src_bands=src_bands, ref_bands=ref_bands, force=force_match,
            ) as raster_fuse:  # yapf: disable
                # construct output filenames
                post_fix = utils.create_out_postfix(
                    raster_fuse.proc_crs, model=model, kernel_shape=kernel_shape, driver=out_profile['driver'],
                )
                corr_filename = out_path.joinpath(src_filename.stem + post_fix)
                param_filename = utils.create_param_filename(corr_filename) if param_image else None

                start_time = timer()
                raster_fuse.process(
                    corr_filename, Model(model), kernel_shape, param_filename=param_filename, build_ovw=build_ovw,
                    overwrite=overwrite, **config,
                )

            logger.info(f'Completed in {timer() - start_time:.2f} secs')
            comp_files += [src_filename, corr_filename]  # build a list of files to pass to compare

        # compare source and corrected files with reference (invokes compare command with relevant parameters)
        if cmp_file:
            if str(cmp_file) == 'ref':
                cmp_file = ref_file
                cmp_bands = ref_bands if not cmp_bands or not len(cmp_bands) else cmp_bands

            cmp_cfg = {k: kwargs[k] for k, v in RasterCompare.create_config().items() if k in kwargs}
            ctx.invoke(
                compare, src_file=comp_files, ref_file=cmp_file, proc_crs=proc_crs,
                src_bands=[src_bands, None] * len(src_file), ref_bands=cmp_bands, force_match=force_match,
                **cmp_cfg
            )

    except Exception:
        logger.exception('Exception caught during processing.')  # log exception info
        raise click.Abort()


cli.add_command(fuse)


# compare command
@cloup.command(cls=HomonimCommand)
@cloup.argument(
    'src-file', nargs=-1, metavar='IMAGE...', type=click.Path(exists=True, dir_okay=False, path_type=pathlib.Path),
    help='Path(s) to image(s) to compare with :option:`REFERENCE`.'
)
@ref_file_arg
@cloup.option_group("Standard options", src_bands_option, ref_bands_option, output_option)
@cloup.option_group(
    "Advanced options", threads_option, max_block_mem_option, downsampling_option, upsampling_option,
    click.option(
        '-pc', '--proc-crs', type=click.Choice([pc.value for pc in ProcCrs], case_sensitive=False),
        default=ProcCrs.auto.value, show_default=True, help="""The image CRS in which to compare images. 
    \b

    - `auto`: lowest resolution of the source and reference CRS's (recommended). 
    - `src`: source image CRS. 
    - `ref`: reference image CRS.
    """
    ),
    force_match_option,  # yapf: disable
)
def compare(
    src_file: Tuple[pathlib.Path, ...], ref_file: pathlib.Path, src_bands: Tuple[int], ref_bands: Tuple[int],
    output: pathlib.Path, proc_crs: ProcCrs, force_match, **kwargs
):
    """
    Compare image(s) with a reference.

    Report similarity statistics between input image(s) and a reference image.  Typically, this is used to compare the
    before and after accuracy of surface reflectance correction, by comparing source and corrected images with a
    new reference image.

    Reference and input image(s) should be co-located and spectrally similar.  Reference image extents must encompass
    those of the input image(s).

    The reference image should contain bands that are approximate (wavelength) matches to the input image bands.
    Where input and reference images are RGB, or have ``center_wavelength`` metadata, bands are matched automatically.
    Where there are the same number of input and reference bands, and no ``center_wavelength`` metadata, bands are
    assumed to be in matching order.  The :option:`--src-band <homonim-compare --src-band>` and
    :option:`--ref-band <homonim-compare --ref-band>` options allow subsets and ordering of input and reference bands
    to be specified.
    \b

    Examples:
    ---------

    Compare `source.tif` and `corrected.tif` with `reference.tif`::

        homonim compare source.tif corrected.tif reference.tif

    ----
    """

    # build configuration dictionary
    config = RasterCompare.create_config(**kwargs)
    try:
        stats_dict = {}
        # if src_bands comes from compare CLI, convert to list[src_bands, ...] with one element for each source file
        src_bands_list = (
            [src_bands] * len(src_file)
            if not src_bands or all([isinstance(src_band, int) for src_band in src_bands])
            else src_bands
        )  # yapf: disable
        # iterate over source files, comparing with reference
        for src_filename, src_bands in zip(src_file, src_bands_list):
            logger.info(f'\nComparing {src_filename.name}')
            start_time = timer()
            with RasterCompare(
                src_filename, ref_file, proc_crs=proc_crs, src_bands=src_bands, ref_bands=ref_bands, force=force_match,
            ) as raster_compare:  # yapf: disable
                stats_dict[str(src_filename)] = raster_compare.process(**config)
            logger.info(f'Completed in {timer() - start_time:.2f} secs')

        # print a key for the following tables
        logger.info(f'\n\n{raster_compare.schema_table()}')

        # print a results table per source image file
        summ_dict = {}
        for src_filename, im_stats_dict in stats_dict.items():
            logger.info(f'\n\n{str(src_filename)}:\n\n{RasterCompare.stats_table(im_stats_dict)}')
            summ_dict[pathlib.Path(src_filename).name] = im_stats_dict['Mean'].copy()

        # print a summary results table comparing all source files
        if len(summ_dict) > 1:
            logger.info(f'\n\nSummary over bands:\n\n{RasterCompare.stats_table(summ_dict, key_header="file")}')

        if output is not None:
            stats_dict['Reference'] = str(ref_file)
            with open(output, 'w') as file:
                json.dump(stats_dict, file)

    except Exception:
        logger.exception('Exception caught during processing')
        raise click.Abort()


cli.add_command(compare)


@cloup.command(cls=HomonimCommand)
@cloup.argument(
    'param-files', nargs=-1, metavar='PARAM...', type=click.Path(exists=True, dir_okay=False, path_type=pathlib.Path),
    callback=_param_file_cb, help='Path(s) to parameter image(s).'
)
@output_option
def stats(param_files: Tuple[pathlib.Path, ...], output: pathlib.Path):
    """
    Report parameter statistics.

    Report the minimum, maximum, mean etc. values of a parameter image generated with the
    :option:`--param-image <homonim-fuse --param-image>` option of the ``fuse`` command.
    """

    try:
        stats_dict = {}
        meta_dict = {}

        # process parameter file(s), storing results
        for param_filename in param_files:
            logger.info(f'\nProcessing {param_filename.name}')
            with ParamStats(param_filename) as param_stats:
                stats_dict[str(param_filename)] = param_stats.stats()
                meta_dict[str(param_filename)] = param_stats.metadata

        # print a key for the following tables
        logger.info(f'\n\n{param_stats.schema_table}')

        # iterate over stored result(s) and print
        for param_filename, param_dict in stats_dict.items():
            logger.info(f'\n{pathlib.Path(param_filename).name}:\n')
            logger.info(meta_dict[param_filename])
            logger.info(f'Stats:\n\n{ParamStats.stats_table(stats_dict[param_filename])}\n')

        if output is not None:
            with open(output, 'w') as file:
                json.dump(stats_dict, file, allow_nan=True)

    except Exception:
        logger.exception('Exception caught during processing')
        raise click.Abort()


cli.add_command(stats)

##



"""
    Homonim: Correction of aerial and satellite imagery to surface reflectance
    Copyright (C) 2021 Dugal Harris
    Email: dugalh@gmail.com

    This program is free software: you can redistribute it and/or modify
    it under the terms of the GNU Affero General Public License as
    published by the Free Software Foundation, either version 3 of the
    License, or any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU Affero General Public License for more details.

    You should have received a copy of the GNU Affero General Public License
    along with this program.  If not, see <https://www.gnu.org/licenses/>.
"""

from pathlib import Path
from typing import Dict, Tuple

import numpy as np
import pytest
import rasterio as rio
from rasterio.crs import CRS
from rasterio.enums import Resampling
from rasterio.transform import Affine, from_bounds
from rasterio.windows import Window
from rasterio.warp import transform_bounds

from homonim.errors import ImageProfileError, ImageFormatError
from homonim.raster_array import RasterArray


def test_read_only_properties(array_byte, profile_byte):
    """ Test RasterArray read-only properties. """
    basic_ra = RasterArray(array_byte, profile_byte['crs'], profile_byte['transform'], nodata=profile_byte['nodata'])
    assert basic_ra.shape == array_byte.shape
    assert basic_ra.width == array_byte.shape[-1]
    assert basic_ra.height == array_byte.shape[-2]
    assert basic_ra.count == 1 if array_byte.ndim < 3 else array_byte.ndim
    assert basic_ra.nodata == profile_byte['nodata']
    assert basic_ra.transform == profile_byte['transform']
    assert basic_ra.dtype == array_byte.dtype
    assert basic_ra.res == (profile_byte['transform'].a, -profile_byte['transform'].e)
    assert basic_ra.crs == profile_byte['crs']


def test_array_property(array_byte, profile_byte):
    """ Test array get/set. """
    byte_ra = RasterArray(array_byte, profile_byte['crs'], profile_byte['transform'], nodata=profile_byte['nodata'])
    assert (byte_ra.array == array_byte).all()  # test get

    array = array_byte / 2
    byte_ra.array = array
    assert (byte_ra.array == array).all()  # test set with same num bands

    array = np.stack((array, array), axis=0)
    byte_ra.array = array
    assert (byte_ra.array == array).all()  # test set with more bands
    assert byte_ra.count == array.shape[0]


def test_nodata_mask(ra_byte):
    """ Test set/get nodata and mask properties. """
    mask = ra_byte.mask
    ra_byte.nodata = 254
    assert ra_byte.nodata == 254
    assert (ra_byte.mask == mask).all()  # test mask unchanged after setting nodata
    assert (ra_byte.mask_ra.array == mask).all()
    assert ra_byte.mask_ra.transform == ra_byte.transform

    # test mask unchanged after setting stacked array
    ra_byte.array = np.stack((ra_byte.array, ra_byte.array), axis=0)
    assert (ra_byte.mask == mask).all()

    # test altering the mask
    mask[np.divide(mask.shape, 2).astype('int')] = False
    ra_byte.mask = mask
    assert (ra_byte.mask == mask).all()

    # test removing nodata
    ra_byte.nodata = None
    assert ra_byte.mask.all()


def test_array_set_shape(ra_byte):
    """ Test setting array with different rows/cols raises error. """
    with pytest.raises(ValueError):
        ra_byte.array = ra_byte.array.reshape(-1, 1)


def test_from_profile(array_byte, profile_byte):
    """ Test creating raster array from rasterio profile dict. """
    byte_ra = RasterArray.from_profile(array_byte, profile_byte)
    assert (byte_ra.array == array_byte).all()
    assert byte_ra.transform == profile_byte['transform']


def test_from_profile_noarray(profile_byte):
    """ Test creating raster array from rasterio profile dict w/o array. """
    byte_ra = RasterArray.from_profile(None, profile_byte)
    assert (byte_ra.array == byte_ra.nodata).all()


@pytest.mark.parametrize('missing_key', ['crs', 'transform', 'nodata', 'width', 'height', 'count', 'dtype'])
def test_from_profile_missingkey(profile_byte, missing_key: str):
    """ Test an error is raised when creating raster array from a rasterio profile dict that is missing a key. """
    profile = profile_byte
    profile.pop(missing_key)
    with pytest.raises(ImageProfileError):
        RasterArray.from_profile(None, profile)


def test_from_rio_dataset(file_byte):
    """ Test an error is raised when creating raster array from a rasterio profile dict that is missing a key. """
    with rio.open(file_byte, 'r') as ds:
        # check default
        ds_ra = RasterArray.from_rio_dataset(ds)
        assert ds_ra.shape == ds.shape
        assert ds_ra.count == ds.count
        assert ds_ra.nodata == ds.nodata
        assert ds_ra.dtype == RasterArray.default_dtype

        # create boundless raster array that extends beyond ds
        pad = [1, 1]
        indexes = ds.indexes[0]
        window = Window(-pad[1], -pad[0], ds.width + 2 * pad[1], ds.height + 2 * pad[0])
        ds_ra_boundless = RasterArray.from_rio_dataset(ds, indexes=indexes, window=window)
        assert ds_ra_boundless.shape == (window.height, window.width)
        assert ds_ra_boundless.count == 1

        # check boundless array contents and transform against ds_ra
        bounded_win = Window(pad[1], pad[0], ds_ra.width, ds_ra.height)
        assert (ds_ra_boundless.array[bounded_win.toslices()] == ds_ra.array).all()
        test_transform = ds_ra.transform * Affine.translation(-pad[1], -pad[0])
        assert (
            ds_ra_boundless.transform.xoff == test_transform.xoff and
            ds_ra_boundless.transform.yoff == test_transform.yoff
        )


@pytest.mark.parametrize('file, count', [('file_masked', 1), ('file_rgba', 3)])
def test_from_rio_dataset_masked(file: str, count: int, request: pytest.FixtureRequest):
    """ Test creating raster array from nodata and internally masked datasets. """
    file: Path = request.getfixturevalue(file)
    with rio.open(file, 'r') as ds:
        ds_mask = ds.dataset_mask().astype('bool', copy=False)
        ra = RasterArray.from_rio_dataset(ds)
        assert ra.count == count
        assert np.isnan(ra.nodata)
        assert (ra.mask == ds_mask).all()


@pytest.mark.parametrize('pad', [[1, 1], [-1, -1]])
def test_bounded_window_slices(file_byte, pad: Tuple[int, int]):
    """ Test RasterArray.bounded_window_slices() with bounded and boundless windows. """
    with rio.open(file_byte, 'r') as ds:
        window = Window(-pad[1], -pad[0], ds.width + 2 * pad[1], ds.height + 2 * pad[0])
        bounded_win, bounded_slices = RasterArray.bounded_window_slices(ds, window)
        # test that the returned window is bounded by the dataset
        assert (bounded_win.col_off == max(0, -pad[1]) and bounded_win.row_off == max(0, -pad[0]))
        assert (
            bounded_win.width == min(ds.width, ds.width + 2 * pad[1]) and
            bounded_win.height == min(ds.height, ds.height + 2 * pad[0])
        )


def test_slice_to_bounds(ra_byte):
    """ Test RasterArray.slice_to_bounds(). """
    # test valid bounded window
    window = Window(1, 1, ra_byte.width - 2, ra_byte.height - 2)
    bounds = ra_byte.window_bounds(window)
    slice_ra = ra_byte.slice_to_bounds(*bounds)
    assert slice_ra.bounds == pytest.approx(bounds)
    assert (slice_ra.array == ra_byte.array[window.toslices()]).all()

    # test invalid boundless window
    with pytest.raises(ValueError):
        ra_byte.slice_to_bounds(*ra_byte.window_bounds(Window(-1, -1, ra_byte.width, ra_byte.height)))


def test_to_rio_dataset(ra_byte, tmp_path: Path):
    """ Test writing raster array to dataset. """
    ds_filename = tmp_path.joinpath('temp.tif')
    with rio.open(ds_filename, 'w', driver='GTiff', **ra_byte.profile) as ds:
        ra_byte.to_rio_dataset(ds)
    with rio.open(ds_filename, 'r') as ds:
        test_array = ds.read(indexes=1)
    assert (test_array == ra_byte.array).all()


def test_to_rio_dataset_crop(ra_rgb_byte, tmp_path: Path):
    """ Test writing a raster array to a dataset where the dataset & raster array sizes differ. """
    ds_filename = tmp_path.joinpath('temp.tif')
    indexes = [1, 2, 3]
    # crop the raster array and write to full dataset
    crop_window = Window(1, 1, ra_rgb_byte.width - 2, ra_rgb_byte.height - 2)
    crop_ra = ra_rgb_byte.slice_to_bounds(*ra_rgb_byte.window_bounds(crop_window))
    with rio.open(ds_filename, 'w', driver='GTiff', **ra_rgb_byte.profile) as ds:
        crop_ra.to_rio_dataset(ds, indexes=indexes, window=crop_window)
    with rio.open(ds_filename, 'r') as ds:
        test_array = ds.read(indexes=indexes)
    assert (test_array[(np.array(indexes) - 1, *crop_window.toslices())] == crop_ra.array).all()

    # crop the dataset and write in the full raster array
    with rio.open(ds_filename, 'w', driver='GTiff', **crop_ra.profile) as ds:
        ra_rgb_byte.to_rio_dataset(ds, indexes=indexes)
    with rio.open(ds_filename, 'r') as ds:
        test_array = ds.read(indexes=indexes)
    assert (test_array == ra_rgb_byte.array[(np.array(indexes) - 1, *crop_window.toslices())]).all()


def test_to_rio_dataset_exceptions(ra_rgb_byte, tmp_path: Path):
    """ Test possible error conditions when writing a raster array to a dataset. """
    ds_filename = tmp_path.joinpath('temp.tif')
    with rio.open(ds_filename, 'w', driver='GTiff', **ra_rgb_byte.profile) as ds:
        with pytest.raises(ValueError):
            # window lies outside the bounds of raster array
            crop_window = Window(1, 1, ra_rgb_byte.width - 2, ra_rgb_byte.height - 2)
            crop_ra = ra_rgb_byte.slice_to_bounds(*ra_rgb_byte.window_bounds(crop_window))
            boundless_window = Window(-1, -1, ra_rgb_byte.width + 2, ra_rgb_byte.height + 2)
            crop_ra.to_rio_dataset(ds, indexes=[1, 2, 3], window=boundless_window)
        with pytest.raises(ValueError):
            # len(indexes) > number of dataset bands
            ra_rgb_byte.to_rio_dataset(ds, indexes=[1] * (ds.count + 1))
        with pytest.raises(ValueError):
            # indexes outside of valid range
            ra_rgb_byte.to_rio_dataset(ds, indexes=ds.count + 1)

    # dataset and raster array have different CRSs
    profile = ra_rgb_byte.profile
    profile.update(crs=CRS.from_epsg(4326))
    with rio.open(ds_filename, 'w', driver='GTiff', **profile) as ds:
        with pytest.raises(ImageFormatError):
            ra_rgb_byte.to_rio_dataset(ds, indexes=[1, 2, 3])

    # dataset and raster array have different resolutions
    profile = ra_rgb_byte.profile
    profile.update(transform=Affine.identity() * Affine.scale(0.5))
    with rio.open(ds_filename, 'w', driver='GTiff', **profile) as ds:
        with pytest.raises(ImageFormatError):
            ra_rgb_byte.to_rio_dataset(ds, indexes=[1, 2, 3])


def test_reprojection(ra_rgb_byte):
    """ Test raster array re-projection. """
    # reproject to WGS84 with default parameters, assuming ra_rgb_byte is North up
    to_crs = CRS.from_epsg(4326)
    reprj_ra = ra_rgb_byte.reproject(crs=to_crs, resampling=Resampling.nearest)
    assert (reprj_ra.crs == to_crs)
    abs_diff = np.abs(reprj_ra.array[:, reprj_ra.mask] - ra_rgb_byte.array[:, ra_rgb_byte.mask])
    assert abs_diff.mean() == pytest.approx(0, abs=.1)

    # reproject with rescaling to WGS84 using a specified transform & shape
    to_bounds = transform_bounds(ra_rgb_byte.crs, to_crs, *ra_rgb_byte.bounds)
    to_shape = tuple(np.array(ra_rgb_byte.shape) * 2)
    to_transform = from_bounds(*to_bounds, *to_shape[::-1])
    reprj_ra = ra_rgb_byte.reproject(
        crs=to_crs, transform=to_transform, shape=to_shape, resampling=Resampling.bilinear
    )
    assert (reprj_ra.crs == to_crs)
    assert (reprj_ra.transform == to_transform)
    assert (reprj_ra.shape == to_shape)
    assert (reprj_ra.bounds == pytest.approx(to_bounds, abs=1.e-9))
    assert (
        reprj_ra.array[:, reprj_ra.mask].mean() == pytest.approx(ra_rgb_byte.array[:, ra_rgb_byte.mask].mean(), abs=.1)
    )



"""
    Homonim: Correction of aerial and satellite imagery to surface reflectance
    Copyright (C) 2021 Dugal Harris
    Email: dugalh@gmail.com

    This program is free software: you can redistribute it and/or modify
    it under the terms of the GNU Affero General Public License as
    published by the Free Software Foundation, either version 3 of the
    License, or any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU Affero General Public License for more details.

    You should have received a copy of the GNU Affero General Public License
    along with this program.  If not, see <https://www.gnu.org/licenses/>.
"""

import os
from pathlib import Path
from typing import Tuple

import pytest
import rasterio as rio
import yaml
from click.testing import CliRunner
from rasterio.warp import Resampling
from rasterio.vrt import WarpedVRT

from homonim import utils
from homonim.cli import cli
from homonim.enums import ProcCrs, Model
from homonim.fuse import RasterFuse
from tests.conftest import str_contain_no_space, FuseCliParams


@pytest.mark.parametrize(
    'model, kernel_shape', [
        (Model.gain, (1, 1)),
        (Model.gain_blk_offset, (1, 1)),
        (Model.gain_offset, (5, 5)),
    ]
)  # yapf: disable
def test_fuse(
    tmp_path: Path, runner: CliRunner, file_rgb_100cm_float, file_rgb_50cm_float, model: Model,
    kernel_shape: Tuple[int, int],
):
    """ Test fuse cli output with different methods and kernel shapes. """
    ref_file = file_rgb_100cm_float
    src_file = file_rgb_50cm_float
    post_fix = utils.create_out_postfix(ProcCrs.ref, model, kernel_shape, RasterFuse.create_out_profile()['driver'])
    corr_file = tmp_path.joinpath(src_file.stem + post_fix)
    cli_str = f'fuse -m {model.value} -k {kernel_shape[0]} {kernel_shape[1]} -od {tmp_path} {src_file} {ref_file}'
    result = runner.invoke(cli, cli_str.split())
    assert (result.exit_code == 0)
    assert (corr_file.exists())

    with rio.open(src_file, 'r') as src_ds, rio.open(corr_file, 'r') as out_ds:
        assert (out_ds.tags()['FUSE_MODEL'] == model.name)
        assert (out_ds.tags()['FUSE_KERNEL_SHAPE'] == str(kernel_shape))

        src_array = src_ds.read(indexes=src_ds.indexes)
        src_mask = src_ds.dataset_mask().astype('bool', copy=False)
        out_array = out_ds.read(indexes=out_ds.indexes)
        out_mask = out_ds.dataset_mask().astype('bool', copy=False)
        assert (out_mask == src_mask).all()
        assert (out_array[:, out_mask] == pytest.approx(src_array[:, src_mask], abs=.1))


def test_fuse_defaults(runner: CliRunner, default_fuse_cli_params: FuseCliParams):
    """ Test fuse cli works without model or kernel shape arguments. """
    result = runner.invoke(cli, default_fuse_cli_params.cli_str.split())
    assert (result.exit_code == 0)
    assert (default_fuse_cli_params.corr_file.exists())


def test_method_error(runner: CliRunner, default_fuse_cli_params: FuseCliParams):
    """ Test unknown model generates an error. """
    cli_str = default_fuse_cli_params.cli_str + ' -m unk'
    result = runner.invoke(cli, cli_str.split())
    assert (result.exit_code != 0)
    assert ("Invalid value for '-m' / '--model'" in result.output)


@pytest.mark.parametrize('bad_kernel_shape', [(0, 0), (2, 3), (3, 2)])
def test_kernel_shape_error(runner: CliRunner, default_fuse_cli_params: FuseCliParams, bad_kernel_shape):
    """ Test bad kernel shape generates an error. """
    cli_str = default_fuse_cli_params.cli_str + f' -k {bad_kernel_shape[0]} {bad_kernel_shape[1]}'
    result = runner.invoke(cli, cli_str.split())
    assert (result.exit_code != 0)
    assert ("kernel_shape" in result.output)


def test_file_exists_error(runner: CliRunner, basic_fuse_cli_params: FuseCliParams):
    """ Test that attempting to overwrite an existing output file generates an error. """
    basic_fuse_cli_params.corr_file.touch()
    cli_str = basic_fuse_cli_params.cli_str
    result = runner.invoke(cli, cli_str.split())
    assert (result.exit_code != 0)
    assert ('FileExistsError' in result.output)

    os.remove(basic_fuse_cli_params.corr_file)
    basic_fuse_cli_params.param_file.touch()
    cli_str = basic_fuse_cli_params.cli_str + ' --param-image'
    result = runner.invoke(cli, cli_str.split())
    assert (result.exit_code != 0)
    assert ('FileExistsError' in result.output)


def test_overwrite(runner: CliRunner, basic_fuse_cli_params: FuseCliParams):
    """ Test overwriting existing output file(s) with -o. """
    basic_fuse_cli_params.corr_file.touch()
    basic_fuse_cli_params.param_file.touch()
    cli_str = basic_fuse_cli_params.cli_str + ' --param-image -o'
    result = runner.invoke(cli, cli_str.split())
    assert (result.exit_code == 0)
    assert (basic_fuse_cli_params.corr_file.exists())
    assert (basic_fuse_cli_params.param_file.exists())


def test_compare(runner: CliRunner, ref_file_100cm_float, src_file_100cm_float):
    """ Test --compare, in flag and value configurations, against expected output. """
    ref_file = ref_file_100cm_float
    src_file = src_file_100cm_float
    # test --compare in flag (no value), and value configuration
    cli_strs = [
        f'fuse  {src_file} {ref_file} --compare', f'fuse {src_file} {ref_file} --compare {ref_file_100cm_float} -o'
    ]
    for cli_str in cli_strs:
        result = runner.invoke(cli, cli_str.split())
        assert (result.exit_code == 0)
        src_cmp_str = """float_100cm_src.tif:
           Band    r²   RMSE   rRMSE   N
    ----------- ----- ------ ------- ---
    Ref. band 1 1.000  0.000   0.000 144
           Mean 1.000  0.000   0.000 144"""
        assert (str_contain_no_space(src_cmp_str, result.output))

        corr_cmp_str = """float_100cm_src_FUSE_cREF_mGAIN-BLK-OFFSET_k5_5.tif:
           Band      r²   RMSE   rRMSE   N
    ----------- ----- ------ ------- ---
    Ref. band 1 1.000  0.000   0.000 144
           Mean 1.000  0.000   0.000 144"""
        assert (str_contain_no_space(corr_cmp_str, result.output))

        sum_cmp_str = """File    r²   RMSE   rRMSE   N
    --------------------------------------------------- ----- ------ ------- ---
                                    float_100cm_src.tif 1.000  0.000   0.000 144
    float_100cm_src_FUSE_cREF_mGAIN-BLK-OFFSET_k5_5.tif 1.000  0.000   0.000 144"""
        assert (str_contain_no_space(sum_cmp_str, result.output))


def test_compare_file_exists_error(runner: CliRunner, ref_file_100cm_float, src_file_100cm_float):
    """ Test --compare raises an exception when the specified file does not exist. """
    ref_file = ref_file_100cm_float
    src_file = src_file_100cm_float
    # test --compare in flag (no value), and value configurayion
    cli_str = f'fuse  {src_file} {ref_file} --compare unknown.tif'
    result = runner.invoke(cli, cli_str.split())
    assert (result.exit_code != 0)
    assert ('No such file or directory' in result.output)


@pytest.mark.parametrize('proc_crs', [ProcCrs.auto, ProcCrs.ref, ProcCrs.src])
def test_proc_crs(
    tmp_path: Path, runner: CliRunner, ref_file_100cm_float, src_file_100cm_float, proc_crs: ProcCrs,
):
    """ Test valid --proc-crs settings generate an output with correct metadata. """
    ref_file = ref_file_100cm_float
    src_file = src_file_100cm_float
    model = Model.gain_blk_offset
    kernel_shape = (3, 3)
    res_proc_crs = ProcCrs.ref if proc_crs == ProcCrs.auto else proc_crs
    post_fix = utils.create_out_postfix(res_proc_crs, model, kernel_shape, RasterFuse.create_out_profile()['driver'])
    corr_file = tmp_path.joinpath(src_file.stem + post_fix)
    cli_str = (
        f'fuse -m {model.value} -k {kernel_shape[0]} {kernel_shape[1]} -od {tmp_path} -pc {proc_crs.value} '
        f'{src_file} {ref_file}'
    )
    result = runner.invoke(cli, cli_str.split())
    assert (result.exit_code == 0)
    assert (corr_file.exists())

    with rio.open(corr_file, 'r') as out_ds:
        assert (out_ds.tags()['FUSE_PROC_CRS'] == res_proc_crs.name)


def test_conf_file(tmp_path: Path, runner: CliRunner, basic_fuse_cli_params: FuseCliParams):
    """ Test passing a configuration file results in a correctly configured output. """
    # create test configuration file
    conf_dict = dict(
        mask_partial=True, param_image=True, dtype='uint8', nodata=0, creation_options=dict(compress='lzw')
    )
    conf_file = tmp_path.joinpath('conf.yaml')
    with open(conf_file, 'w') as f:
        yaml.dump(conf_dict, f)

    cli_str = basic_fuse_cli_params.cli_str + f' -c {conf_file}'
    result = runner.invoke(cli, cli_str.split())
    assert (result.exit_code == 0)
    assert (basic_fuse_cli_params.corr_file.exists())
    assert (basic_fuse_cli_params.param_file.exists())  # test param_image==True

    with rio.open(basic_fuse_cli_params.src_file, 'r') as src_ds:
        with rio.open(basic_fuse_cli_params.corr_file, 'r') as out_ds:
            # test nodata, dtype and creation_options
            assert (out_ds.nodata == conf_dict['nodata'])
            assert (out_ds.dtypes[0] == conf_dict['dtype'])
            assert (out_ds.profile['compress'] == conf_dict['creation_options']['compress'])
            # test mask_partial==True
            src_mask = src_ds.dataset_mask().astype('bool', copy=False)
            out_mask = out_ds.dataset_mask().astype('bool', copy=False)
            assert (src_mask[out_mask].all())
            assert (src_mask.sum() > out_mask.sum())
            # test proc_crs
            assert (out_ds.tags()['FUSE_PROC_CRS'] == basic_fuse_cli_params.proc_crs.name)


def test_param_image(runner: CliRunner, basic_fuse_cli_params: FuseCliParams):
    """ Test --param-image. """
    # test that cli without --param-image generates no parameter image
    cli_str = basic_fuse_cli_params.cli_str
    result = runner.invoke(cli, cli_str.split())
    assert (result.exit_code == 0)
    assert (basic_fuse_cli_params.corr_file.exists())
    assert (not basic_fuse_cli_params.param_file.exists())

    # test --param-image generates a valid parameter image
    cli_str = basic_fuse_cli_params.cli_str + ' --param-image -o'
    result = runner.invoke(cli, cli_str.split())
    assert (result.exit_code == 0)
    assert (basic_fuse_cli_params.corr_file.exists())
    assert (basic_fuse_cli_params.param_file.exists())
    utils.validate_param_image(basic_fuse_cli_params.param_file)


def test_mask_partial(runner: CliRunner, basic_fuse_cli_params: FuseCliParams):
    """ Test --mask-partial. """
    cli_str = basic_fuse_cli_params.cli_str + ' --mask-partial'
    result = runner.invoke(cli, cli_str.split())
    assert (result.exit_code == 0)
    assert (basic_fuse_cli_params.corr_file.exists())

    with rio.open(basic_fuse_cli_params.src_file, 'r') as src_ds:
        with rio.open(basic_fuse_cli_params.corr_file, 'r') as out_ds:
            # test that the output mask is contained by and smaller than the src mask
            src_mask = src_ds.dataset_mask().astype('bool', copy=False)
            out_mask = out_ds.dataset_mask().astype('bool', copy=False)
            assert (src_mask[out_mask].all())
            assert (src_mask.sum() > out_mask.sum())


def test_threads(runner: CliRunner, basic_fuse_cli_params: FuseCliParams):
    """ Test --threads. """
    cli_str = basic_fuse_cli_params.cli_str + ' --threads 1'
    result = runner.invoke(cli, cli_str.split())
    assert (result.exit_code == 0)
    assert (basic_fuse_cli_params.corr_file.exists())


def test_max_block_mem(runner: CliRunner, basic_fuse_cli_params: FuseCliParams):
    """ Test --max-block-mem. """
    cli_str = basic_fuse_cli_params.cli_str + ' -mbm 1e-4'
    result = runner.invoke(cli, cli_str.split())
    assert (result.exit_code == 0)
    assert (basic_fuse_cli_params.corr_file.exists())

    # test that max_block_mem too small raises a BlockSizeError
    cli_str = basic_fuse_cli_params.cli_str + ' -o -mbm 1e-6'
    result = runner.invoke(cli, cli_str.split())
    assert (result.exit_code != 0)
    assert ('BlockSizeError' in result.output)


@pytest.mark.parametrize('upsampling', [r.name for r in rio.warp.SUPPORTED_RESAMPLING])
def test_upsampling(runner: CliRunner, basic_fuse_cli_params: FuseCliParams, upsampling: Resampling):
    """ Test --upsampling with valid values generates output with correct metadata. """
    cli_str = basic_fuse_cli_params.cli_str + f' --upsampling {upsampling}'
    result = runner.invoke(cli, cli_str.split())
    assert (result.exit_code == 0)
    assert (basic_fuse_cli_params.corr_file.exists())
    with rio.open(basic_fuse_cli_params.corr_file, 'r') as out_ds:
        tags_dict = out_ds.tags()
        assert ('FUSE_UPSAMPLING' in tags_dict)
        assert (yaml.safe_load(tags_dict['FUSE_UPSAMPLING']) == upsampling)


@pytest.mark.parametrize('downsampling', [r.name for r in rio.warp.SUPPORTED_RESAMPLING])
def test_downsampling(runner: CliRunner, basic_fuse_cli_params: FuseCliParams, downsampling: Resampling):
    """ Test --downsampling with valid values generates output with correct metadata. """
    cli_str = basic_fuse_cli_params.cli_str + f' --downsampling {downsampling}'
    result = runner.invoke(cli, cli_str.split())
    assert (result.exit_code == 0)
    assert (basic_fuse_cli_params.corr_file.exists())
    with rio.open(basic_fuse_cli_params.corr_file, 'r') as out_ds:
        tags_dict = out_ds.tags()
        assert ('FUSE_DOWNSAMPLING' in tags_dict)
        assert (yaml.safe_load(tags_dict['FUSE_DOWNSAMPLING']) == downsampling)


def test_upsampling_error(runner: CliRunner, basic_fuse_cli_params: FuseCliParams):
    """ Test --upsampling with bad value raises an error. """
    cli_str = basic_fuse_cli_params.cli_str + f' --upsampling unknown'
    result = runner.invoke(cli, cli_str.split())
    assert (result.exit_code != 0)
    assert ("Invalid value for '-us' / '--upsampling'" in result.output)


def test_downsampling_error(runner: CliRunner, basic_fuse_cli_params: FuseCliParams):
    """ Test --downsampling with bad value raises an error. """
    cli_str = basic_fuse_cli_params.cli_str + f' --downsampling unknown'
    result = runner.invoke(cli, cli_str.split())
    assert (result.exit_code != 0)
    assert ("Invalid value for '-ds' / '--downsampling'" in result.output)


@pytest.mark.parametrize('r2_inpaint_thresh', [0, 0.5, 1])
def test_r2_inpaint_thresh(runner: CliRunner, basic_fuse_cli_params: FuseCliParams, r2_inpaint_thresh: float):
    """ Test --r2-inpaint-thresh generates an output with correct metadata. """
    cli_str = basic_fuse_cli_params.cli_str + f' --r2-inpaint-thresh {r2_inpaint_thresh}'
    result = runner.invoke(cli, cli_str.split())
    assert (result.exit_code == 0)
    assert (basic_fuse_cli_params.corr_file.exists())
    with rio.open(basic_fuse_cli_params.corr_file, 'r') as out_ds:
        tags_dict = out_ds.tags()
        assert ('FUSE_R2_INPAINT_THRESH' in tags_dict)
        assert (yaml.safe_load(tags_dict['FUSE_R2_INPAINT_THRESH']) == r2_inpaint_thresh)


@pytest.mark.parametrize('bad_r2_inpaint_thresh', [-1, 2])
def test_r2_inpaint_thresh_error(runner: CliRunner, basic_fuse_cli_params: FuseCliParams, bad_r2_inpaint_thresh: float):
    """ Test --r2-inpaint-thresh with bad value raises an error. """
    cli_str = basic_fuse_cli_params.cli_str + f' --r2-inpaint-thresh {bad_r2_inpaint_thresh}'
    result = runner.invoke(cli, cli_str.split())
    assert (result.exit_code != 0)
    assert ('Invalid value' in result.output)


@pytest.mark.parametrize(
    'driver, dtype, nodata', [
        ('GTiff', 'float64', float('nan')),
        ('GTiff', 'uint16', 65535),
        ('PNG', 'uint8', 0),
    ]
)  # yapf: disable
def test_out_profile(runner: CliRunner, basic_fuse_cli_params: FuseCliParams, driver: str, dtype: str, nodata: float):
    """ Test --out-* options generate a correctly configured output. """
    cli_str = basic_fuse_cli_params.cli_str + f' --driver {driver} --dtype {dtype} --nodata {nodata}'
    ext_dict = rio.drivers.raster_driver_extensions()
    ext_idx = list(ext_dict.values()).index(driver)
    ext = list(ext_dict.keys())[ext_idx]
    corr_file = basic_fuse_cli_params.corr_file.parent.joinpath(f'{basic_fuse_cli_params.corr_file.stem}.{ext}')
    result = runner.invoke(cli, cli_str.split())
    assert (result.exit_code == 0)
    assert (corr_file.exists())
    with rio.open(corr_file, 'r') as out_ds:
        assert (out_ds.driver == driver)
        assert (out_ds.dtypes[0] == dtype)
        assert (utils.nan_equals(out_ds.nodata, nodata))


def test_out_driver_error(runner: CliRunner, basic_fuse_cli_params: FuseCliParams):
    """ Test --driver with invalid value raises an error. """
    cli_str = basic_fuse_cli_params.cli_str + f' --driver unk'
    result = runner.invoke(cli, cli_str.split())
    assert (result.exit_code != 0)
    assert ('Invalid value' in result.output)


def test_out_dtype_error(runner: CliRunner, basic_fuse_cli_params: FuseCliParams):
    """ Test --dtype with invalid value raises an error. """
    cli_str = basic_fuse_cli_params.cli_str + f' --dtype unk'
    result = runner.invoke(cli, cli_str.split())
    assert (result.exit_code != 0)
    assert ('Invalid value' in result.output)


def test_out_nodata_error(runner: CliRunner, basic_fuse_cli_params: FuseCliParams):
    """ Test --nodata with invalid value (cannot be cast to --dtype) raises an error. """
    cli_str = basic_fuse_cli_params.cli_str + f' --dtype uint8 --nodata nan'
    result = runner.invoke(cli, cli_str.split())
    assert (result.exit_code != 0)
    assert ('Invalid value' in result.output)


def test_creation_options(runner: CliRunner, basic_fuse_cli_params: FuseCliParams):
    """ Test -co creation options generate correctly configured output. """
    cli_str = basic_fuse_cli_params.cli_str + f' -co COMPRESS=LZW -co PREDICTOR=2 -co TILED=NO'
    result = runner.invoke(cli, cli_str.split())
    assert (result.exit_code == 0)
    assert (basic_fuse_cli_params.corr_file.exists())
    with rio.open(basic_fuse_cli_params.corr_file, 'r') as out_ds:
        assert (out_ds.profile['compress'] == 'lzw')
        assert (not out_ds.profile['tiled'])


@pytest.mark.parametrize(
    'src_bands, ref_bands, force, exp_bands', [
        ((3, 2, 1), None, False, (3, 2, 1)),
        ((2, 1), (3, 1, 2), False, (2, 1)),
        ((2, 1), (3, 2, 1), True, (3, 2)),
    ]
)  # yapf: disable
def test_src_ref_bands(
    src_bands: Tuple[int], ref_bands: Tuple[int], force: bool, exp_bands: Tuple[int],
    default_fuse_rgb_cli_params: FuseCliParams, tmp_path: Path, runner: CliRunner,
):
    """ Test fuse with --src_band, --ref_band and --force-match parameters. """
    cli_str = default_fuse_rgb_cli_params.cli_str
    if src_bands:
        cli_str += ''.join([' -sb ' + str(bi) for bi in src_bands])
    if ref_bands:
        cli_str += ''.join([' -rb ' + str(bi) for bi in ref_bands])
    if force:
        cli_str += ' -f'

    result = runner.invoke(cli, cli_str.split())
    assert (result.exit_code == 0)
    assert (default_fuse_rgb_cli_params.corr_file.exists())
    with WarpedVRT(rio.open(default_fuse_rgb_cli_params.src_file, 'r')) as src_ds:
        with rio.open(default_fuse_rgb_cli_params.corr_file, 'r') as out_ds:
            src_array = src_ds.read(indexes=exp_bands)
            src_mask = src_ds.dataset_mask().astype('bool', copy=False)
            out_array = out_ds.read()
            out_mask = out_ds.dataset_mask().astype('bool', copy=False)

            assert out_ds.count == len(exp_bands)
            assert (out_mask == src_mask).all()
            assert (out_array[:, out_mask] == pytest.approx(src_array[:, src_mask], abs=2))

@pytest.mark.parametrize(
    'src_bands, ref_bands, cmp_bands, force, cmp_ref, exp_bands', [
        (None, None, None, False, False, (1, 2, 3)),
        ((1, 2), (1, 2), (1, 2), False, False, (1, 2)),
        ((1, 2), (1, 2), None, True, False, (1, 2)),
        ((3, 1), (3, 1, 2), None, True, True, (3, 1)),
    ]
)  # yapf: disable
def test_src_ref_cmp_bands(
    src_bands: Tuple[int], ref_bands: Tuple[int], cmp_bands: Tuple[int], force: bool, cmp_ref: bool,
    exp_bands: Tuple[int], default_fuse_rgb_cli_params: FuseCliParams, tmp_path: Path, runner: CliRunner,
):
    """ Test fuse --compare with --src_band, --ref_band, --force-match and --cmp-band parameters. """
    # When bands are matched based on assumed RGB center wavelengths, the corrected file is (intentionally) not written
    # with center wavelengths.  Depending on how --src-band is spec'd, This can result in corrected files with < 3
    # bands, or corrected files with bands not in RGB order.  This in turn can lead to problems with --compare,
    # where these kinds of corrected files cannot be matched with the compare reference, or are matched incorrectly.
    # The above parameter cases avoid any of these situations.  It doesn't seem possible to work around this without
    # writing assumed RGB center wavelengths to corrected files, which seems like a bad idea.  Rather I just
    # generate a warning when RGB wavelengths are assumed.  Perhaps I should also always print out how
    # bands are matched?  Practically, I think most of the time people will correct RGB->RGB, so we wouldn't see this
    # issue often at all.

    cli_str = default_fuse_rgb_cli_params.cli_str
    if src_bands:
        cli_str += ''.join([' -sb ' + str(bi) for bi in src_bands])
    if ref_bands:
        cli_str += ''.join([' -rb ' + str(bi) for bi in ref_bands])
    if force:
        cli_str += ' -f'
    cli_str += ' -cmp' if cmp_ref else f' -cmp {str(default_fuse_rgb_cli_params.ref_file)}'
    if cmp_bands:
        cli_str += ''.join([' -cb ' + str(bi) for bi in cmp_bands])

    result = runner.invoke(cli, cli_str.split())
    assert (result.exit_code == 0)
    assert (default_fuse_rgb_cli_params.corr_file.exists())

    with WarpedVRT(rio.open(default_fuse_rgb_cli_params.src_file, 'r')) as src_ds:
        with rio.open(default_fuse_rgb_cli_params.corr_file, 'r') as out_ds:
            src_array = src_ds.read(indexes=exp_bands)
            src_mask = src_ds.dataset_mask().astype('bool', copy=False)
            out_array = out_ds.read()
            out_mask = out_ds.dataset_mask().astype('bool', copy=False)

            assert out_ds.count == len(exp_bands)
            assert (out_mask == src_mask).all()
            assert (out_array[:, out_mask] == pytest.approx(src_array[:, src_mask], abs=2))

    test_str = """                                              File    r²   RMSE   rRMSE   N
-------------------------------------------------- ----- ------ ------- ---
                                float_50cm_rgb.tif 1.000  0.000   0.000 144
float_50cm_rgb_FUSE_cREF_mGAIN-BLK-OFFSET_k5_5.tif 1.000  0.000   0.000 144"""
    assert test_str in result.output



"""
    Homonim: Correction of aerial and satellite imagery to surface reflectance
    Copyright (C) 2021 Dugal Harris
    Email: dugalh@gmail.com

    This program is free software: you can redistribute it and/or modify
    it under the terms of the GNU Affero General Public License as
    published by the Free Software Foundation, either version 3 of the
    License, or any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU Affero General Public License for more details.

    You should have received a copy of the GNU Affero General Public License
    along with this program.  If not, see <https://www.gnu.org/licenses/>.
"""
import logging
from pathlib import Path
from typing import List, Tuple, Union, Dict
import warnings

import numpy as np
import rasterio as rio
from rasterio.enums import ColorInterp
from tabulate import tabulate
from homonim.raster_pair import RasterPairReader
from homonim.enums import ProcCrs

logger = logging.getLogger(__name__)

class MatchedPairReader(RasterPairReader):
    _max_rel_wavelength_diff = .1   # maximum allowed relative distance between center wavelengths for matching

    def __init__(
        self, src_filename: Union[str, Path], ref_filename: Union[str, Path], proc_crs: ProcCrs = ProcCrs.auto,
        src_bands: Tuple[int, ...] = None, ref_bands: Tuple[int, ...] = None, force: bool = False,
    ):
        """
        Class for reading matching, and optionally overlapping, blocks from a source and reference image pair.

        Reference and source image(s) should be co-located and spectrally similar.  Reference image extents must
        encompass those of the source image.

        The reference image should contain bands that are approximate (wavelength) matches to the source image bands.
        Where source and reference images are RGB, or have ``center_wavelength`` metadata, bands are matched
        automatically based on wavelength.  Where there are the same number of source and reference bands, and no
        ``center_wavelength`` metadata, bands are assumed to be in matching order.  Subsets and ordering of source
        and reference bands can be specified with the ``src_bands`` and ``ref_bands`` parameters.

        .. note::

            Images downloaded with `geedim <https://github.com/leftfield-geospatial/geedim>`_ have ``center_wavelength`` metadata
            compatible with ``homonim``.

        Parameters
        ----------
        src_filename: str, pathlib.Path
            Path or URL of the source image file.
        ref_filename: str, pathlib.Path
            Path or URL of the reference image file.
        proc_crs: homonim.enums.ProcCrs, optional
            :class:`~homonim.enums.ProcCrs` instance specifying which of the source/reference image CRS and pixel
            grid to use for processing.  For most use cases, it can be left as the default of
            :attr:`~homonim.enums.ProcCrs.auto` i.e. the lowest resolution of the source and reference image CRS's.
        src_bands: list of int, optional.
            Indexes of source spectral bands to be processed (1 based).  If not specified, all bands with the
            ``center_wavelength`` property, or all non-alpha bands, are used.
        ref_bands: list of int, optional.
            Indexes of reference spectral bands to match and fuse with source bands (1 based).  Should contain at
            least as many elements as ``src_bands``, or the number of valid bands in the source image file,
            if ``src_bands`` is not specified.  If ``ref_bands`` is not specified, all reference bands with the
            ``center_wavelength`` property, or all non-alpha bands are used.
        force: bool, optional
            Bypass auto wavelength matching, and any band-matching errors.  Use with caution.
        """
        self._src_bands = src_bands
        self._ref_bands = ref_bands
        self._force = force
        RasterPairReader.__init__(self, src_filename, ref_filename, proc_crs=proc_crs)

    @property
    def src_bands(self) -> Tuple[int, ...]:
        """ Matched source band indices (1-based). """
        return self._src_bands

    @property
    def ref_bands(self) ->Tuple[int, ...]:
        """ Matched reference band indices (1-based). """
        return self._ref_bands

    @staticmethod
    def _get_band_info(
        im: rio.DatasetReader, bands: Tuple[int, ...] = None, name: str = None
    ) -> Tuple[np.ndarray, np.ndarray, np.ndarray]:
        """ Return band indices, corresponding names, and center wavelengths (if any). """
        name = name or Path(im.name).name or ''
        non_alpha_bands = np.array([
            i + 1 for i in range(im.count)
            if (
                (im.colorinterp[i] != ColorInterp.alpha) and not
                (im.descriptions[i] and (im.descriptions[i].endswith('_MASK') or im.descriptions[i].endswith('_DIST')))
            )   # if not alpha band or geedim mask band
        ])  # yapf: disable
        refl_bands = np.array([bi for bi in non_alpha_bands if 'center_wavelength' in im.tags(bi)])

        # test bands are valid
        if (bands is not None) and (not set(bands).issubset(range(1, im.count + 1))):
            invalid_bands = list(set(bands).difference(range(1, im.count + 1)))
            raise ValueError(f'User specified {name} bands contain invalid band(s) {invalid_bands}.')
        # test bands do not contain alpha bands
        if (bands is not None) and (not set(bands).issubset(non_alpha_bands)):
            alpha_bands = list(set(bands).difference(non_alpha_bands))
            raise ValueError(f'User specified {name} bands contain alpha band(s) {alpha_bands}.')
        # warn if some but not all bands have center wavelength metadata
        if (bands is not None) and len(refl_bands) and (not set(bands).issubset(refl_bands)):
            non_refl_bands = list(set(bands).difference(refl_bands))
            logger.warning(f'User specified {name} bands contain non-reflectance band index(es) {non_refl_bands}.')

        log_prefix = f'Using {name} bands '
        if (bands is not None) and (len(bands) > 0):
            # use bands if it was specified
            bands = np.array(bands)
        elif len(refl_bands) > 0:
            # else use bands with center wavelengths if any
            log_prefix = f'Using {name} reflectance bands '
            bands = np.array(refl_bands)
        elif len(non_alpha_bands) > 0:
            # else use non-alpha bands
            log_prefix = f'Using {name} non-alpha bands '
            bands = np.array(non_alpha_bands)
        else:
            raise ValueError(f'There are no non-alpha/reflectance {name} bands to use.')

        # center wavelengths for all bands
        center_wavelengths = np.array([
            float(im.tags(bi)['center_wavelength'])
            if 'center_wavelength' in im.tags(bi) else np.nan
            for bi in range(1, im.count + 1)
        ])  # yapf: disables

        # assign standard RGB center wavelengths
        if (len(non_alpha_bands) == 3):
            std_rgb_cws = dict(zip(
                [ColorInterp.red, ColorInterp.green, ColorInterp.blue], [.650, .560, .480]
            ))  # yapf: disable
            log_list = []
            # assign standard RGB center wavelengths according to valid colorinterp's
            for bi in non_alpha_bands:
                if not np.isnan(center_wavelengths[bi - 1]):
                    continue    # don't overwrite existing center wavelengths
                if im.colorinterp[bi - 1] in std_rgb_cws:
                    center_wavelengths[bi - 1] = std_rgb_cws[im.colorinterp[bi - 1]]
                    log_list.append(im.colorinterp[bi - 1].name)
            if len(log_list) > 0:
                logger.warning(
                    f'Assigning standard {", ".join(log_list)} center wavelengths for {name}.'
                )

            # if the image has no valid RGB colorinterp's or center wavelengths
            if sum(np.isnan(center_wavelengths[non_alpha_bands - 1])) == 3:
                # assume RGB and assign center wavelengths in that order
                logger.warning(
                    f'Assuming image is RGB, and assigning standard center wavelengths for {name}.'
                )
                center_wavelengths[non_alpha_bands - 1] = list(std_rgb_cws.values())

        center_wavelengths = center_wavelengths[bands - 1]
        band_names = np.array([im.descriptions[bi - 1] if im.descriptions[bi - 1] else str(bi) for bi in bands])
        logger.debug(f'{log_prefix} {list(band_names)}.')
        return bands, band_names, center_wavelengths

    def _get_pair_band_table(
        self, src_im: rio.DatasetReader, ref_im: rio.DatasetReader, src_bands: Tuple[int, ...] = None,
        ref_bands: Tuple[int, ...] = None
    ):
        """ Return a table of source and reference metadata for the given bands. """
        num_bands = min(len(src_bands), len(ref_bands))
        src_bands = src_bands[:num_bands]
        ref_bands = ref_bands[:num_bands]

        # retrieve src/ref image band metadata as lists of dicts
        src_band_list = []
        ref_band_list = []
        band_keys = {'name': 'Name', 'description': 'Description', 'center_wavelength': 'Wavelen'}
        for band_list, im, bands in zip([src_band_list, ref_band_list], [src_im, ref_im], [src_bands, ref_bands]):
            for band in bands:
                band_dict = {}
                if 'name' not in im.tags(band):
                    band_dict.update(Name=im.descriptions[band-1] or str(band))
                band_tags = im.tags(band)
                band_dict.update(**{bkn: band_tags[bk] for bk, bkn in band_keys.items() if bk in band_tags})
                band_list.append(band_dict)

        # combine src and ref lists of dicts into one
        def prefix_dict_keys(band_dict: Dict, prefix: str):
            return {f'{prefix}\n{k}':v for k, v in band_dict.items()}

        band_list = []
        super_keys = dict()
        for src_band_dict, ref_band_dict in zip(src_band_list, ref_band_list):
            band_dict = {}
            band_dict.update(prefix_dict_keys(src_band_dict, 'Source'))
            band_dict.update(prefix_dict_keys(ref_band_dict, 'Ref'))
            band_list.append(band_dict)
            super_keys.update(dict.fromkeys(band_dict.keys()))

        # update missing keys with empty string vals
        for band_dict in band_list:
            update_keys = [k for k in super_keys if k not in band_dict.keys()]
            update_vals = [''] * len(update_keys)
            band_dict.update(zip(update_keys, update_vals))

        return tabulate(band_list, headers='keys',  maxcolwidths=20, tablefmt='simple')

    def _match_pair_bands(
        self, src_im: rio.DatasetReader, ref_im: rio.DatasetReader
    ) -> Tuple[Tuple[int], Tuple[int]]:  # yapf: disable
        """
        Validate and match source and reference bands. An override of base class RasterPairReader._match_pair_bands().
        """
        # retrieve non-alpha bands
        src_bands, src_band_names, src_wavelengths = MatchedPairReader._get_band_info(src_im, bands=self._src_bands)
        ref_bands, ref_band_names, ref_wavelengths = MatchedPairReader._get_band_info(ref_im, bands=self._ref_bands)
        src_name = Path(src_im.name).name
        ref_name = Path(ref_im.name).name

        if len(src_bands) > len(ref_bands):
            if not self._force:
                raise ValueError(f'{ref_name} has fewer bands than {src_name}.')
            else:
                logger.warning(f'{ref_name} has fewer bands than {src_name}.')

        match_bands = np.array([np.nan] * len(src_bands))
        # match src with ref bands based on center wavelength metadata, bypassing if force==True
        if any(src_wavelengths) and any(ref_wavelengths) and not self._force:
            # TODO: consider using a linear programming type optimisation here,
            #  e.g. https://stackoverflow.com/questions/67368093/find-optimal-unique-neighbour-pairs-based-on-closest-distance

            # absolute & relative distance matrix between src and ref center wavelengths
            abs_dist = np.abs(src_wavelengths[:, np.newaxis] - ref_wavelengths[np.newaxis, :])
            rel_dist = abs_dist / src_wavelengths[:, np.newaxis]
            def greedy_match(dist: np.ndarray) -> Tuple[np.ndarray, np.ndarray]:
                """
                Greedy matching of src to ref bands based on the provided center wavelength distance matrix,
                `dist`. src bands must be down the rows, and ref bands along the cols of `dist`.
                Will match one ref band for each src band until either all src or all ref bands
                have been matched.  Works for all cases where len(src_bands) != len(ref_bands).
                """
                # match_idx[i] is the index of the ref band that matches with the ith src band
                match_idx = np.array([np.nan] * dist.shape[0])
                match_dist = np.array([np.nan] * dist.shape[0]) # distances corresponding to the above matches

                # suppress runtime warning on all-Nan slice, as it is expected in normal operation
                with warnings.catch_warnings():
                    warnings.simplefilter('ignore', category=RuntimeWarning)
                    # repeat until all src or ref bands have been matched
                    while not all(np.isnan(np.nanmin(dist, axis=1))) or not all(np.isnan(np.nanmin(dist, axis=0))):
                        # find the row with the smallest distance in it
                        min_dist = np.nanmin(dist, axis=1)
                        min_dist_row_idx = np.nanargmin(min_dist)
                        min_dist_row = dist[min_dist_row_idx, :]
                        # store match idx and distance for this row
                        match_idx[min_dist_row_idx] = np.nanargmin(min_dist_row)
                        match_dist[min_dist_row_idx] = min_dist[min_dist_row_idx]
                        # set the matched row and col to nan, so that it is not used again
                        dist[:, int(match_idx[min_dist_row_idx])] = np.nan
                        dist[min_dist_row_idx, :] = np.nan

                return match_dist, match_idx

            match_dist, match_idx = greedy_match(rel_dist)

            # if any of the matched distances are greater than a threshold, raise an informative error,
            # or log a warning, depending on `self._force`
            if any(match_dist > MatchedPairReader._max_rel_wavelength_diff):
                err_idx = match_dist > MatchedPairReader._max_rel_wavelength_diff
                src_err_band_names = list(src_band_names[err_idx])
                ref_err_band_names = list(ref_band_names[np.int64(match_idx[err_idx])])
                err_dists = list(match_dist[err_idx].round(3))
                raise ValueError(
                    f'{src_name} band(s) {src_err_band_names} could not be auto-matched.  The nearest '
                    f'{ref_name} band(s) were {ref_err_band_names}, at center wavelength difference(s) of '
                    f'{err_dists} (um) respectively.'
                )

            src_matched = ~np.isnan(match_idx)
            match_bands[src_matched] = ref_bands[np.int64(match_idx[src_matched])]
            logger.debug(
                f'Matching {src_name} band(s) {list(src_band_names[src_matched])} with {ref_name} band(s) '
                f'{list(ref_band_names[np.int64(match_idx[src_matched])])}, at center wavelength difference(s) of '
                f'{list(match_dist[src_matched].round(3))} (um) respectively.'
            )

        # match any remaining bands that don't have center wavelength metadata
        if sum(~np.isnan(match_bands)) < min(len(src_bands), len(ref_bands)):
            unmatched = np.isnan(match_bands)
            unmatch_ref_idx = np.array([i for i, bi in enumerate(ref_bands) if bi not in match_bands], dtype=int)
            unmatch_ref_bands = ref_bands[unmatch_ref_idx]
            unmatch_ref_band_names = ref_band_names[unmatch_ref_idx]
            if len(src_bands) == len(ref_bands):
                # assume unmatched src and ref image bands are in matching order
                match_bands[unmatched] = unmatch_ref_bands
                logger.debug(
                    f'Matching {src_name} band(s) {list(src_band_names[unmatched])} in file order with'
                    f' {ref_name} band(s) {list(unmatch_ref_band_names)}.'
                )
            elif self._force:
                # match the remaining N src bands with the first N unmatched ref bands (N=sum(unmatched))
                unmatch_ref_bands = unmatch_ref_bands[:sum(unmatched)]
                unmatch_ref_band_names = unmatch_ref_band_names[:sum(unmatched)]
                # if there are not N unmatched ref bands, truncate unmatched to just match what we can
                unmatched = np.where(unmatched)[0][:len(unmatch_ref_bands)]
                match_bands[unmatched] = unmatch_ref_bands
                logger.warning(
                    f'Matching {src_name} band(s) {list(src_bands[unmatched])} in file order with {ref_name} '
                    f'band(s): {list(unmatch_ref_band_names)}.'
                )
            else:
                # raise an error when remaining unmatched bands counts do not match and `self._force` is False
                raise ValueError(
                    f'Could not match {src_name} band(s) {list(src_bands[unmatched])} with {ref_name} '
                    f'band(s) {list(unmatch_ref_band_names)}.  Ensure {src_name} and {ref_name} non-alpha band '
                    f'counts match, {src_name} and {ref_name} have ``center_wavelength`` tags for each band, '
                    f'or set `force` to True.'
                )
        # Truncate src and ref to reflect the matched bands
        src_matched = ~np.isnan(match_bands)
        src_bands = src_bands[src_matched]
        ref_matched = np.array([list(ref_bands).index(bi) for bi in match_bands[src_matched]])
        ref_bands = ref_bands[ref_matched]
        logger.debug('Matched band(s):\n\n' + self._get_pair_band_table(src_im, ref_im, src_bands, ref_bands) + '\n')
        return tuple(src_bands.tolist()), tuple(ref_bands.tolist())



"""
    Homonim: Correction of aerial and satellite imagery to surface reflectance
    Copyright (C) 2021 Dugal Harris
    Email: dugalh@gmail.com

    This program is free software: you can redistribute it and/or modify
    it under the terms of the GNU Affero General Public License as
    published by the Free Software Foundation, either version 3 of the
    License, or any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU Affero General Public License for more details.

    You should have received a copy of the GNU Affero General Public License
    along with this program.  If not, see <https://www.gnu.org/licenses/>.
"""
from enum import Enum


class Model(str, Enum):
    """
    Linear model variants for correcting to surface reflectance.

    Roughly speaking, gain compensates for atmospheric absorption and anisotropic (BRDF) effects, and offset
    (when present) compensates for atmospheric reflectance and haze.
    """
    gain = 'gain'
    """ 
    Gain-only model, suitable for haze-free and zero offset images (i.e. images where a surface reflectance of 
    zero corresponds to a pixel value of ± zero). 
    """
    gain_blk_offset = 'gain-blk-offset'
    """ 
    Gain-only model applied to offset normalised image blocks.  Suitable for most source-reference combinations.
    """
    gain_offset = 'gain-offset'
    """
    Gain and offset model.  The most accurate model, but sensitive to differences between source and reference, 
    such as shadowing and land cover changes.  Suitable for well-matched source / reference image pairs.  
    """


class ProcCrs(str, Enum):
    """
    CRS and pixel grid in which images will be processed.
    """
    auto = 'auto'
    """ Lowest resolution of the source and reference image CRS's (recommended). """
    src = 'src'
    """ Source image CRS. """
    ref = 'ref'
    """ Reference image CRS. """



"""
    Homonim: Correction of aerial and satellite imagery to surface reflectance
    Copyright (C) 2021 Dugal Harris
    Email: dugalh@gmail.com

    This program is free software: you can redistribute it and/or modify
    it under the terms of the GNU Affero General Public License as
    published by the Free Software Foundation, either version 3 of the
    License, or any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU Affero General Public License for more details.

    You should have received a copy of the GNU Affero General Public License
    along with this program.  If not, see <https://www.gnu.org/licenses/>.
"""

from pathlib import Path
from typing import Tuple

import numpy as np
import pytest
import rasterio as rio
from pytest import FixtureRequest
from rasterio import MemoryFile
from rasterio.enums import Resampling
from rasterio.windows import Window, union
from rasterio.errors import RasterioIOError

from homonim.enums import ProcCrs
from homonim.errors import ImageContentError, BlockSizeError, IoError
from homonim.raster_pair import RasterPairReader


@pytest.mark.parametrize(
    'src_file, ref_file, expected_proc_crs', [
        ('src_file_50cm_float', 'ref_file_100cm_float', ProcCrs.ref),
        ('src_file_100cm_float', 'ref_file_50cm_float', ProcCrs.src)
    ]
)  # yapf: disable
def test_creation(src_file: str, ref_file: str, expected_proc_crs: ProcCrs, request: FixtureRequest):
    """ Test RasterPair creation and proc_crs resolution. """
    src_file: Path = request.getfixturevalue(src_file)
    ref_file: Path = request.getfixturevalue(ref_file)
    raster_pair = RasterPairReader(src_file, ref_file)
    assert (raster_pair.proc_crs == expected_proc_crs)
    assert (raster_pair.src_bands == (1, ))
    assert (raster_pair.ref_bands == (1, ))

    # enter the raster pair context and test block(s) correspond to bands
    with raster_pair as rp:
        block_pairs = list(rp.block_pairs())
        assert (len(block_pairs) == 1)
        assert (block_pairs[0].src_out_block == Window(0, 0, rp.src_im.width, rp.src_im.height))
    assert rp.closed


@pytest.mark.parametrize(
    'src_file, ref_file', [
        ('ref_file_50cm_float', 'src_file_100cm_float'),
        ('ref_file_50cm_float', 'src_file_50cm_float'),
        ('ref_file_100cm_float', 'src_file_50cm_float'),
        ('ref_file_100cm_float', 'src_file_100cm_float')
    ]
)  # yapf: disable
def test_coverage_exception(src_file: str, ref_file: str, request: FixtureRequest):
    """ Test that ref not covering the extent of src raises an error. """
    src_file: Path = request.getfixturevalue(src_file)
    ref_file: Path = request.getfixturevalue(ref_file)
    with pytest.raises(ImageContentError):
        _ = RasterPairReader(src_file, ref_file)


def test_band_count_exception(file_rgba, file_byte):
    """ Test that src band count > ref band count raises an error. """
    with pytest.raises(ImageContentError):
        _ = RasterPairReader(file_rgba, file_byte)


def test_block_shape_exception(src_file_50cm_float, ref_file_100cm_float):
    """ Test block shape errors. """
    # test auto block shape smaller than a pixel
    raster_pair = RasterPairReader(src_file_50cm_float, ref_file_100cm_float)
    with pytest.raises(BlockSizeError):
        with raster_pair:
            block_pairs = [block_pair for block_pair in raster_pair.block_pairs(max_block_mem=1.e-5)]

    # test auto block shape smaller than overlap
    with pytest.raises(BlockSizeError):
        with raster_pair:
            block_pairs = [block_pair for block_pair in raster_pair.block_pairs(overlap=(5, 5), max_block_mem=1.e-4)]


def test_not_open_exception(src_file_50cm_float, ref_file_100cm_float):
    """ Test that using a raster pair before entering the context raises an error. """
    raster_pair = RasterPairReader(src_file_50cm_float, ref_file_100cm_float)
    with pytest.raises(IoError):
        _ = raster_pair.ref_im
    with pytest.raises(IoError):
        _ = raster_pair.src_im
    with pytest.raises(IoError):
        _ = list(raster_pair.block_pairs())
    with pytest.raises(IoError):
        _ = raster_pair.read(None)


@pytest.mark.parametrize(
    'src_file, ref_file, proc_crs, blk_overlap, max_block_mem', [
        ('src_file_45cm_float', 'ref_file_100cm_float', ProcCrs.auto, (0, 0), 1.e-3),
        ('src_file_45cm_float', 'ref_file_100cm_float', ProcCrs.auto, (2, 2), 1.e-3),
        ('src_file_45cm_float', 'ref_file_100cm_float', ProcCrs.auto, (0, 0), 2.e-4),
        ('src_file_45cm_float', 'ref_file_100cm_float', ProcCrs.auto, (1, 1), 2.e-4),
        ('src_file_100cm_float', 'ref_file_45cm_float', ProcCrs.auto, (0, 0), 1.e-3),
        ('src_file_100cm_float', 'ref_file_45cm_float', ProcCrs.auto, (2, 2), 1.e-3),
        ('src_file_100cm_float', 'ref_file_45cm_float', ProcCrs.auto, (0, 0), 2.e-4),
        ('src_file_100cm_float', 'ref_file_45cm_float', ProcCrs.auto, (1, 1), 2.e-4),
    ]
)  # yapf: disable
def test_block_pair_continuity(
    src_file: str, ref_file: str, proc_crs: ProcCrs, blk_overlap: Tuple[int, int], max_block_mem: float,
    request: FixtureRequest
):
    """ Test the continuity of block pairs for different src/ref etc combinations. """
    src_file: Path = request.getfixturevalue(src_file)
    ref_file: Path = request.getfixturevalue(ref_file)
    raster_pair = RasterPairReader(src_file, ref_file, proc_crs=proc_crs)

    def compare_blocks(block: Window, prev_block: Window, overlap: Tuple[int, int] = (0, 0), compare=np.equal):
        """ Test block continuity. """
        if block.row_off == prev_block.row_off:  # blocks in the same row
            assert compare(block.col_off, prev_block.col_off + prev_block.width - 2 * overlap[1])
        else:
            assert compare(block.row_off, prev_block.row_off + prev_block.height - 2 * overlap[0])

    with raster_pair:
        # Create lists of compare_blocks() parameters for each block in a BlockPair.
        # NOTE: the <'other' crs>_in_block may overlap more than ``overlap`, otherwise <proc_crs>_in_block's should
        # overlap by exactly blk_overlap, and *out_blocks should be exactly adjacent.  max_block_mem, and blk_overlap
        # should however be chosen to give a block shape > 2*overlap.
        block_keys = ['src_in_block', 'ref_in_block', 'src_out_block', 'ref_out_block']
        # *in_blocks overlap, *out_blocks don't
        overlaps = [blk_overlap, blk_overlap, (0, 0), (0, 0)]
        if raster_pair.proc_crs == ProcCrs.ref:
            # the src_in_block can overlap by more than blk_overlap, the other blocks should be exact
            compares = [np.less_equal, np.equal, np.equal, np.equal]
        else:
            # the ref_in_block can overlap by more than blk_overlap, the other blocks should be exact
            compares = [np.equal, np.less_equal, np.equal, np.equal]
        block_pairs = list(raster_pair.block_pairs(overlap=blk_overlap, max_block_mem=max_block_mem))
        prev_block_pair = block_pairs[0]
        for block_pair in block_pairs[1:]:
            if block_pair.band_i == prev_block_pair.band_i:  # blocks in the same band
                # compare each block type with its previous version, using the appropriate compare_block params
                for block_key, overlap, compare in zip(block_keys, overlaps, compares):
                    block = block_pair.__getattribute__(block_key)
                    prev_block = prev_block_pair.__getattribute__(block_key)
                    compare_blocks(block, prev_block, overlap, compare)
            else:
                assert (block_pair.band_i == prev_block_pair.band_i + 1)
            prev_block_pair = block_pair


@pytest.mark.parametrize(
    'src_file, ref_file, proc_crs, overlap, max_block_mem', [
        ('src_file_45cm_float', 'ref_file_100cm_float', ProcCrs.auto, (0, 0), 1.e-3),
        ('src_file_45cm_float', 'ref_file_100cm_float', ProcCrs.auto, (2, 2), 1.e-3),
        ('src_file_45cm_float', 'ref_file_100cm_float', ProcCrs.auto, (0, 0), 2.e-4),
        ('src_file_45cm_float', 'ref_file_100cm_float', ProcCrs.auto, (2, 2), 2.e-4),
        ('src_file_100cm_float', 'ref_file_45cm_float', ProcCrs.auto, (0, 0), 1.e-3),
        ('src_file_100cm_float', 'ref_file_45cm_float', ProcCrs.auto, (2, 2), 1.e-3),
        ('src_file_100cm_float', 'ref_file_45cm_float', ProcCrs.auto, (0, 0), 2.e-4),
        ('src_file_100cm_float', 'ref_file_45cm_float', ProcCrs.auto, (2, 2), 2.e-4),
    ]
)  # yapf: disable
def test_block_pair_coverage(
    src_file: str, ref_file: str, proc_crs: ProcCrs, overlap: Tuple[int, int], max_block_mem: float,
    request: FixtureRequest,
):
    """ Test that combined block pairs cover the processing window for different src/ref etc combinations. """
    src_file: Path = request.getfixturevalue(src_file)
    ref_file: Path = request.getfixturevalue(ref_file)
    raster_pair = RasterPairReader(src_file, ref_file, proc_crs=proc_crs)
    with raster_pair:
        block_pairs = list(raster_pair.block_pairs(overlap=overlap, max_block_mem=max_block_mem))
        accum_block_pair = block_pairs[0]._asdict()  # a dict to hold combined windows

        # find the combined windows for the block pairs
        for block_pair in block_pairs[1:]:
            for field in ['src_in_block', 'ref_in_block', 'src_out_block', 'ref_out_block']:
                accum_block_pair[field] = union(block_pair.__getattribute__(field), accum_block_pair[field])

        # test coverage of the combined windows
        if raster_pair.proc_crs == ProcCrs.ref:
            assert (accum_block_pair['ref_in_block'] == raster_pair._ref_win)
            assert (accum_block_pair['ref_out_block'] == raster_pair._ref_win)
            src_win = Window(0, 0, raster_pair.src_im.width, raster_pair.src_im.height)
            assert (accum_block_pair['src_in_block'].intersection(src_win) == src_win)
            assert (accum_block_pair['src_out_block'].intersection(src_win) == src_win)
        elif raster_pair.proc_crs == ProcCrs.src:
            assert (accum_block_pair['src_in_block'] == raster_pair._src_win)
            assert (accum_block_pair['src_out_block'] == raster_pair._src_win)
            assert (accum_block_pair['ref_in_block'].intersection(raster_pair._ref_win) == raster_pair._ref_win)
            assert (accum_block_pair['ref_out_block'].intersection(raster_pair._ref_win) == raster_pair._ref_win)

@pytest.mark.parametrize(
    'src_file, ref_file, proc_crs, overlap, max_block_mem', [
        ('src_file_45cm_float', 'ref_file_100cm_float', ProcCrs.auto, (0, 0), 1.e-3),
        ('src_file_45cm_float', 'ref_file_100cm_float', ProcCrs.auto, (2, 2), 1.e-3),
        ('src_file_45cm_float', 'ref_file_100cm_float', ProcCrs.auto, (0, 0), 2.e-4),
        ('src_file_45cm_float', 'ref_file_100cm_float', ProcCrs.auto, (2, 2), 2.e-4),
        ('src_file_100cm_float', 'ref_file_45cm_float', ProcCrs.auto, (0, 0), 1.e-3),
        ('src_file_100cm_float', 'ref_file_45cm_float', ProcCrs.auto, (2, 2), 1.e-3),
        ('src_file_100cm_float', 'ref_file_45cm_float', ProcCrs.auto, (0, 0), 2.e-4),
        ('src_file_100cm_float', 'ref_file_45cm_float', ProcCrs.auto, (2, 2), 2.e-4),
    ]
)  # yapf: disable
def test_block_pair_io(
    src_file: str, ref_file: str, proc_crs: ProcCrs, overlap: Tuple[int, int], max_block_mem: float,
    request: FixtureRequest,
):
    """
    Test block pairs can be read, reprojected and written as raster arrays without loss of data.

    This is more an integration test with raster array than a raster pair unit test.  It simulates the way raster
    arrays are reprojected in *KernelModel.
    """
    src_file: Path = request.getfixturevalue(src_file)
    ref_file: Path = request.getfixturevalue(ref_file)
    raster_pair = RasterPairReader(src_file, ref_file, proc_crs=proc_crs)

    if raster_pair.proc_crs == ProcCrs.ref:
        ref_sampling = Resampling.average
        src_sampling = Resampling.cubic_spline
    else:
        ref_sampling = Resampling.cubic_spline
        src_sampling = Resampling.average

    # test re-projections from src->ref->src and ref->src->ref for both proc_crs=ref & src
    for reproj_ra in ['src', 'ref']:
        # create src and ref test datasets for writing, and enter the raster pair context
        with MemoryFile() as src_mf, MemoryFile() as ref_mf, raster_pair:
            # create profiles that don't use the VRT driver, which can't be written to
            src_profile = raster_pair.src_im.meta.copy()
            src_profile.update(driver='GTiff')
            ref_profile = raster_pair.ref_im.meta.copy()
            ref_profile.update(driver='GTiff')
            with src_mf.open(**src_profile) as src_ds, ref_mf.open(**ref_profile) as ref_ds:   # yapf: disable
                # read, reproject and write block pairs to their respective datasets
                block_pairs = list(raster_pair.block_pairs(overlap=overlap, max_block_mem=max_block_mem))
                for block_pair in block_pairs:
                    src_ra, ref_ra = raster_pair.read(block_pair)
                    if reproj_ra == 'src':
                        _src_ra = src_ra.reproject(**ref_ra.proj_profile, resampling=ref_sampling)
                        __src_ra = _src_ra.reproject(**src_ra.proj_profile, resampling=src_sampling)
                        _src_ra.to_rio_dataset(ref_ds, indexes=1, window=block_pair.ref_out_block)
                        __src_ra.to_rio_dataset(src_ds, indexes=1, window=block_pair.src_out_block)
                    else:
                        _ref_ra = ref_ra.reproject(**src_ra.proj_profile, resampling=src_sampling)
                        __ref_ra = _ref_ra.reproject(**ref_ra.proj_profile, resampling=ref_sampling)
                        __ref_ra.to_rio_dataset(ref_ds, indexes=1, window=block_pair.ref_out_block)
                        _ref_ra.to_rio_dataset(src_ds, indexes=1, window=block_pair.src_out_block)

            # test the written datasets contain same valid areas as the original src/ref files
            with rio.open(src_file, 'r') as src_ds, src_mf.open() as test_src_ds:
                src_mask = src_ds.read_masks(indexes=1).astype('bool', copy=False)
                test_mask = test_src_ds.read_masks(indexes=1).astype('bool', copy=False)
                assert (test_mask[src_mask]).all()

            with rio.open(ref_file, 'r') as ref_ds, ref_mf.open() as test_ref_ds:
                ref_mask = ref_ds.read_masks(indexes=1).astype('bool', copy=False)
                test_mask = test_ref_ds.read_masks(indexes=1).astype('bool', copy=False)
                assert (test_mask[ref_mask]).all()


@pytest.mark.parametrize(
    'src_file, ref_file, proc_crs', [
        ('src_file_100cm_float', 'ref_file_100cm_float', ProcCrs.ref),
        ('src_file_sup_100cm_float', 'ref_file_100cm_float', ProcCrs.ref),
        ('src_file_rot_100cm_float', 'ref_file_100cm_float', ProcCrs.ref),
        ('src_file_wgs84_100cm_float', 'ref_file_100cm_float', ProcCrs.ref),
        ('src_file_wgs84_sup_100cm_float', 'ref_file_100cm_float', ProcCrs.ref),
        ('src_file_wgs84_100cm_float', 'ref_file_100cm_float', ProcCrs.src),
        ('src_file_wgs84_sup_100cm_float', 'ref_file_100cm_float', ProcCrs.src),
        ('src_file_100cm_float', 'ref_file_sup_100cm_float', ProcCrs.ref),
        ('src_file_100cm_float', 'ref_file_rot_100cm_float', ProcCrs.ref),
        ('src_file_100cm_float', 'ref_file_wgs84_100cm_float', ProcCrs.ref),
        ('src_file_100cm_float', 'ref_file_wgs84_sup_float', ProcCrs.ref),
        ('src_file_100cm_float', 'ref_file_wgs84_100cm_float', ProcCrs.src),
        ('src_file_100cm_float', 'ref_file_wgs84_sup_float', ProcCrs.src),
    ]
)  # yapf: disable
def test_orientation_crs(src_file: str, ref_file: str, proc_crs: ProcCrs, request: FixtureRequest):
    """ Test reference and source in different CRS's and orientations. """
    src_file: Path = request.getfixturevalue(src_file)
    ref_file: Path = request.getfixturevalue(ref_file)
    with RasterPairReader(src_file, ref_file, proc_crs=proc_crs) as raster_pair:
        block_pairs = list(raster_pair.block_pairs())
        for block_pair in block_pairs:
            src_ra, ref_ra = raster_pair.read(block_pair)
            _src_ra = src_ra.reproject(**ref_ra.proj_profile, resampling=Resampling.bilinear)
            abs_diff = np.abs(ref_ra.array - _src_ra.array)
            assert np.nanmean(abs_diff) == pytest.approx(0, abs=1)
            assert np.all(_src_ra.mask[ref_ra.mask])
            _ref_ra = ref_ra.reproject(**src_ra.proj_profile, resampling=Resampling.bilinear)
            abs_diff = np.abs(src_ra.array - _ref_ra.array)
            assert np.nanmean(abs_diff) == pytest.approx(0, abs=1)
            assert np.all(_ref_ra.mask[src_ra.mask])


def test_url():
    """ Test source and reference as URLs rather than files. """
    modis_url = 'https://raw.githubusercontent.com/leftfield-geospatial/homonim/main/tests/data/reference/modis_nbar.tif'
    with RasterPairReader(modis_url, modis_url) as raster_pair:
        assert not raster_pair.closed


def test_file_exists_exception(ngi_src_file, landsat_ref_file):
    """ Test an error is raised if src/ref file doesn't exist. """
    for src_file, ref_file in zip(
        [ngi_src_file, 'dummy.tif'], ['dummy.tif', landsat_ref_file],
    ):
        with pytest.raises(RasterioIOError) as ex:
            with RasterPairReader(src_file, ref_file) as raster_pair:
                pass
        assert 'No such file or directory' in str(ex)
